# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QRect, QSize, Qt
from qtpy.QtWidgets import QHBoxLayout, QLabel, QStackedLayout, QWidget

import karabogui.access as krb_access
from karabo.common.api import State
from karabogui.binding.api import ProxyStatus
from karabogui.indicators import STATE_COLORS, get_device_status_pixmap
from karabogui.request import send_property_changes
from karabogui.util import generateObjectName
from karabogui.widgets.hints import KaraboSceneWidget

from .utils import get_proxy, is_controller_enabled


class ControllerContainer(KaraboSceneWidget, QWidget):
    """A container for scene widgets which provides the interface needed
    by the scene tools.
    """

    def __init__(self, klass, model, parent):
        super().__init__(model=model, parent=parent)
        # Variables used by editable widgets
        self.is_editable = False
        self._style_sheet = ""

        self.status_symbol = QLabel("", self)
        self.status_symbol.setAttribute(Qt.WA_TransparentForMouseEvents)

        self.layout = QStackedLayout(self)
        self.layout.setStackingMode(QStackedLayout.StackAll)
        self.layout.addWidget(self.status_symbol)

        proxies = [get_proxy(*key.split(".", 1)) for key in self.model.keys]
        self.widget_controller = self._create_controller(klass, proxies)
        self._setup_wrapped_widget()
        self.setGeometry(QRect(model.x, model.y, model.width, model.height))
        # Trigger the status change once (might be offline)
        proxy = self.widget_controller.proxy.root_proxy
        self._proxy_status_changed("status", proxy.status)

    def add_proxies(self, proxies):
        """Add proxies to a controller which allows more than one proxy.

        :returns: `True` if successful, otherwise `False`
        """
        controller = self.widget_controller
        proxies_added = []
        for proxy in proxies:
            if not controller.visualize_additional_property(proxy):
                continue
            proxies_added.append(proxy)
            if self.is_editable:
                proxy.on_trait_change(
                    self._on_user_edit, "edit_value,binding.config_update")
                self.setEditableToolTip(proxy.binding.requiredAccessLevel)
            else:
                self.setToolTip(", ".join(self.model.keys))
        return len(proxies_added) == len(proxies)

    def remove_proxies(self, proxies):
        """Remove proxies from a controller which allows more than one proxy.

        :returns: `True` if successful, otherwise `False`
        """
        controller = self.widget_controller
        proxies_removed = []
        for proxy in proxies:
            if not controller.remove_additional_property(proxy):
                continue
            proxies_removed.append(proxy)
            if self.is_editable:
                proxy.on_trait_change(
                    self._on_user_edit, "edit_value,binding.config_update",
                    remove=True)
                self.setEditableToolTip(proxy.binding.requiredAccessLevel)
            else:
                self.setToolTip(", ".join(self.model.keys))
        return len(proxies_removed) == len(proxies)

    def apply_changes(self):
        """Apply user-entered values to the remote device properties
        """
        if not (self.is_editable and self.widget_controller.proxy.visible):
            return

        send_property_changes(self.widget_controller.proxies)

    def decline_changes(self):
        """Undo any user-entered changes
        """
        if not (self.is_editable and self.widget_controller.proxy.visible):
            return

        for proxy in self.widget_controller.proxies:
            proxy.revert_edit()
        # allow the controller to react on the decline
        self.widget_controller.on_decline()

    def destroy(self):
        """Tell the controller to clean up"""
        proxy = self.widget_controller.proxy
        proxy.on_trait_change(self._proxy_status_changed, "existing",
                              remove=True)
        proxy.root_proxy.on_trait_change(self._proxy_status_changed, "status",
                                         remove=True)
        if self.is_editable:
            for proxy in self.widget_controller.proxies:
                proxy.on_trait_change(
                    self._on_user_edit, "edit_value,binding.config_update",
                    remove=True)
            if proxy.binding is None:
                proxy.on_trait_change(
                    self._proxy_binding_changed, "binding.schema_update",
                    remove=True)
        self.widget_controller.destroy()
        self.widget_controller = None
        super().destroy()

    def set_geometry(self, rect):
        self.model.trait_set(x=rect.x(), y=rect.y(),
                             width=rect.width(), height=rect.height())
        self.setGeometry(rect)

    def set_visible(self, visible):
        """Set whether this widget is seen by the user."""
        if visible:
            self.widget_controller.show()
        else:
            self.widget_controller.hide()

    def translate(self, offset):
        new_pos = self.pos() + offset
        self.model.trait_set(x=new_pos.x(), y=new_pos.y())
        self.move(new_pos)

    def update_global_access_level(self, level):
        """Enable the widget based on a new global access level."""
        proxy = self.widget_controller.proxy
        binding = proxy.binding
        if binding is not None:
            enabled = is_controller_enabled(proxy, level)
            self.widget_controller.setEnabled(enabled)
            if self.is_editable:
                self.setEditableToolTip(binding.requiredAccessLevel)

    # ---------------------------------------------------------------------
    # Qt methods

    def keyPressEvent(self, event):
        """Watch for editing events involving the enter and escape keys
        """
        key_code = event.key()
        if key_code == Qt.Key_Escape:
            self.decline_changes()
        elif key_code in (Qt.Key_Enter, Qt.Key_Return):
            self.apply_changes()
        super().keyPressEvent(event)

    # ---------------------------------------------------------------------
    # Internal methods

    def _create_controller(self, klass, proxies):
        """Create the binding controller instance.
        """
        controller = klass(proxy=proxies[0], model=self.model)
        controller.create(self)
        for proxy in proxies[1:]:
            # XXX: Assert that this returns True??
            controller.visualize_additional_property(proxy)

        # Attach a handler for the "status" trait of our main device
        proxy = controller.proxy
        proxy.on_trait_change(self._proxy_status_changed, "existing")
        proxy.root_proxy.on_trait_change(self._proxy_status_changed, "status")
        return controller

    def setEditableToolTip(self, level):
        """Set a tooltip for an editable controller"""
        keys = ", ".join(self.model.keys)
        allowed = krb_access.GLOBAL_ACCESS_LEVEL >= level
        tooltip = f"AccessLevel: {level.name} - Access: {allowed}\n\n{keys}"
        self.setToolTip(tooltip)

    def _proxy_binding_changed(self, proxy, name, new):
        """Traits notification callback when the binding of the proxy changes.

        This is set to synchronize the access level of the widget
        """
        proxy.on_trait_change(self._proxy_binding_changed,
                              "binding.schema_update", remove=True)
        self.setEditableToolTip(proxy.binding.requiredAccessLevel)

    def _proxy_status_changed(self, name, value):
        """Traits notification callback when the status of the proxy changes.

        The existing property of the proxy is evaluated here as it is coupled
        to the schema status change!
        """
        proxy = self.widget_controller.proxy
        if name == "status":
            existing = proxy.existing
            status = value
            if status is ProxyStatus.OFFLINE:
                self.widget_controller.clear_widget()
        elif name == "existing":
            existing = value
            status = proxy.root_proxy.status
        if not existing:
            status = ProxyStatus.MISSING

        pixmap = get_device_status_pixmap(status)
        if pixmap is not None:
            self.status_symbol.setPixmap(pixmap)
            self.status_symbol.show()
        else:
            self.status_symbol.hide()

    def _setup_wrapped_widget(self):
        """Wrap up the alarm symbol and possible edit buttons in a layout with
        the `widget_controller`.
        """
        controller = self.widget_controller
        layout = ControllerLayout()
        layout.addWidget(controller.widget)

        # create our color layout
        layout_widget = QWidget()
        layout_widget.setLayout(layout)
        objectName = generateObjectName(self)
        layout_widget.setObjectName(objectName)
        self._style_sheet = (f"QWidget#{objectName}" +
                             " {{ background-color : rgba{}; }}")

        proxies = controller.proxies
        if self.model.parent_component == "EditableApplyLaterComponent":
            self.is_editable = True
            for proxy in proxies:
                proxy.on_trait_change(
                    self._on_user_edit, "edit_value,binding.config_update")
            layout.setContentsMargins(2, 1, 2, 1)
        else:
            layout.setContentsMargins(0, 0, 1, 1)

        # Tell the widget if it's editing
        controller.set_read_only(not self.is_editable)
        if self.is_editable:
            binding = proxies[0].binding
            if binding is None:
                proxy.on_trait_change(self._proxy_binding_changed,
                                      "binding.schema_update")
                # Use normal tooltip and wait for update
                self.setToolTip(", ".join(self.model.keys))
            else:
                self.setEditableToolTip(level=binding.requiredAccessLevel)
        else:
            self.setToolTip(", ".join(self.model.keys))

        # Make the widget show something
        controller.finish_initialization()
        self.layout.addWidget(layout_widget)

    # ---------------------------------------------------------------------
    # Editing related code

    def _on_user_edit(self, binding, name, value):
        """Trait notification handler for when the value of a PropertyProxy
        changes from user or external action.
        """
        # Do some trait name filtering
        if name not in ("edit_value", "config_update"):
            return

        self._update_background_color()

    def _update_background_color(self):
        proxy = self.widget_controller.proxy
        if not self.is_editable or proxy.binding is None:
            return

        proxies = self.widget_controller.proxies
        if any(proxy.edit_value is not None for proxy in proxies):
            color = STATE_COLORS[State.CHANGING] + (128,)
        else:
            # XXX: We make sure not to lose our stylesheet and apply a
            # transparent color!
            color = (0, 0, 0, 0)

        formatted_sheet = self._style_sheet.format(color)
        self.setStyleSheet(formatted_sheet)


class ControllerLayout(QHBoxLayout):
    """This is the layout for the layout widget that contains the controller
       widget.

       There is a need to subclass the minimum size to not dynamically update
       the minimum size depending on the content margins (for the borders)."""

    def minimumSize(self):
        size = super().minimumSize()
        if not size.isEmpty():
            left, top, right, bottom = self.getContentsMargins()
            size = QSize(size.width() - (left + right),
                         size.height() - (top + bottom))
        return size



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import TraitError, Undefined

from karabo.common import const
from karabo.native import (
    AccessMode, Assignment, Hash, Timestamp, has_changes, is_equal)

from .binding_types import (
    BindingNamespace, BindingRoot, NodeBinding, SlotBinding, VectorHashBinding)
from .proxy import PropertyProxy
from .validate import sanitize_table_value


def iterate_binding(node, base=""):
    """Flat iterate over the binding namespace to yield `key` and `binding`

    Note: Remove recursive binding iteration in future
    """
    namespace = node.value
    base = base + "." if base else ""
    for name in namespace:
        subname = base + name
        subnode = getattr(namespace, name)
        if isinstance(subnode, NodeBinding):
            yield from iterate_binding(subnode, base=subname)
        elif not isinstance(subnode, SlotBinding):
            yield subname, subnode


def apply_fast_data(config, binding, timestamp):
    """Recursively set values from a fast data Hash object to a binding
    object.
    """
    namespace = binding.value
    assert isinstance(namespace, BindingNamespace)

    for key, value, _ in config.iterall():
        if key not in namespace:
            continue

        node = getattr(namespace, key)
        if isinstance(value, Hash) and isinstance(node, NodeBinding):
            apply_fast_data(value, node, timestamp)
        else:
            traits = {"value": value, "timestamp": timestamp}
            # Set everything at once and notify via the config_update event
            try:
                node.trait_set(trait_change_notify=False, **traits)
            except TraitError:
                # value in the configuration is not compatible to schema
                continue
            node.config_update = True

    binding.config_update = True


def apply_configuration(config, binding):
    """Recursively set values from a configuration Hash object to a binding
    object.

    This method sets an `online` (network) configuration Hash to a proxy
    binding and thus always notifies and does not include attributes!
    """
    namespace = binding.value
    assert isinstance(namespace, BindingNamespace)

    for key, value, attrs in config.iterall():
        if key not in namespace:
            continue

        node = getattr(namespace, key)
        if isinstance(value, Hash) and isinstance(node, NodeBinding):
            apply_configuration(value, node)
        else:
            # Set the timestamp no matter what
            ts = Timestamp.fromHashAttributes(attrs) or Timestamp()
            traits = {"value": value, "timestamp": ts}
            # Set everything at once and notify via the config_update event
            try:
                node.trait_set(trait_change_notify=False, **traits)
            except TraitError:
                # value in the configuration is not compatible to schema
                continue
            node.config_update = True

    # Notify listeners
    binding.config_update = True


def apply_default_configuration(binding):
    """Recursively set default values for a binding object.
    """
    assert isinstance(binding, BindingRoot)

    for _, node in iterate_binding(binding):
        default = node.attributes.get(const.KARABO_SCHEMA_DEFAULT_VALUE)
        if default is not None:
            node.value = default
        else:
            node.value = Undefined


def apply_project_configuration(config, binding, base=""):
    """Recursively set values from a configuration Hash object to a binding
    object of a project device.

    A project configuration consists of `values` and `attributes` and is
    in certain cases sanitized if necessary (table element).

    Setting a project configuration sets the value directly and does not
    notify!

    :Returns: failure `dict` of keys and error messages
    """
    namespace = binding.value
    assert isinstance(namespace, BindingNamespace)

    fails = Hash()
    base = base + "." if base else ""

    for key, value, attrs in config.iterall():
        if key not in namespace:
            continue

        # For storage in fail dict!
        subkey = base + key
        node = getattr(namespace, key)
        if isinstance(value, Hash) and isinstance(node, NodeBinding):
            fails.update(apply_project_configuration(value, node, base=subkey))
        else:
            if isinstance(node, VectorHashBinding):
                success, value = sanitize_table_value(node, value)
                if not success:
                    fails.update(
                        {subkey: "The table configuration was "
                                 "sanitized - <b>The table is corrupted</b>"})
            traits = {"value": value}
            # Set the timestamp no matter what
            ts = Timestamp.fromHashAttributes(attrs)
            traits["timestamp"] = ts or Timestamp()
            try:
                node.trait_set(trait_change_notify=False, **traits)
            except TraitError:
                # value in the configuration is not compatible to schema
                fails.update({subkey: f"Setting value of {value} failed"})

    return fails


def extract_configuration(binding) -> Hash:
    """Extract all the values set on a binding into a Hash object.
    """
    assert isinstance(binding, BindingRoot)

    config = Hash()
    for key, node in iterate_binding(binding):
        value = node.value
        if value is Undefined:
            continue
        config[key] = node.value

    return config


def extract_edits(schema, binding) -> Hash:
    """Extract all user edited (non default) values on a binding into a
    Hash object.
    """
    assert isinstance(binding, BindingRoot)

    def _is_readonly(key):
        return schema.hash[key, "accessMode"] == AccessMode.READONLY.value

    config = Hash()
    for key, node in iterate_binding(binding):
        default = node.attributes.get(const.KARABO_SCHEMA_DEFAULT_VALUE)
        value = None if node.value is Undefined else node.value

        changes = not is_equal(default, value)
        is_readonly = _is_readonly(key)
        if not changes or is_readonly:
            continue
        config[key] = value

    return config


def extract_online_edits(schema, binding) -> Hash:
    """Extract all user edited (non default) online values on a binding into a
    Hash object.

    This function takes a class schema as reference to extract the online
    edits and compares them to the default of the schema.
    Only values are considered which are different from the default.

    :param schema: The class schema
    :param binding: The binding of the online device

    :returns success: Boolean to indicate if we have a successful online to
                      offline transition, e.g. no DeviceNodes, ChoiceOfNodes,
                      ListOfNodes.

    """
    assert isinstance(binding, BindingRoot)

    def not_writable(key):
        """Check if a property is writable from external"""
        attributes = schema.hash[key, ...]
        read_only = attributes["accessMode"] == AccessMode.READONLY
        internal = attributes["assignment"] == Assignment.INTERNAL
        return read_only or internal

    config = Hash()
    for key, node in iterate_binding(binding):
        # Injected properties are not considered!
        if key not in schema.hash or not_writable(key):
            continue

        value = node.value
        if value is Undefined:
            continue

        # We take the online defaults to account schema injection. If a default
        # value from the online device is_equal, it should not go into the
        # configuration
        default = node.attributes.get(const.KARABO_SCHEMA_DEFAULT_VALUE)
        if not has_changes(default, value):
            continue
        config[key] = value

    return config


def extract_sparse_configurations(proxies, devices=None):
    """Extract values set on the bindings of a list of `PropertyProxy`
    instances into a dictionary of `Hash` objects (one per device).

    If desired, an existing dictionary object can be provided so that a
    configuration can be iteratively accumulated.
    """
    assert all(isinstance(p, PropertyProxy) for p in proxies)

    devices = {} if devices is None else devices
    for proxy in proxies:
        key, binding = proxy.path, proxy.binding
        if binding is None or proxy.edit_value is None:
            continue

        device_id = proxy.root_proxy.device_id
        hsh = devices.setdefault(device_id, Hash())
        hsh[key] = proxy.edit_value

    return devices


def extract_init_configuration(binding, config):
    """Extract an init configuration from `config` with `binding`

    This function will filter out read only properties without runtime
    attributes from the configurations
    """

    ret = Hash()

    for key, node in iterate_binding(binding):
        # key must be in configuration for processing!
        if key in config:
            value, attrs = config.getElement(key)
            # We only want reconfigurable properties, but in Karabo 2.11
            # read only parameter can still have attributes.
            read_only = node.accessMode is AccessMode.READONLY
            is_internal = node.assignment is Assignment.INTERNAL
            if read_only or is_internal:
                continue
            default = node.attributes.get(const.KARABO_SCHEMA_DEFAULT_VALUE)
            if is_equal(default, value):
                continue
            # We make use of and also filter the runtime attributes
            ret[key] = value

    return ret


def get_config_changes(old, new, project):
    """Extract the changes of Hash `new` with respect to Hash `old`

    :param project: Additional parameter denoting if this comparison is
                    happening for a project device proxy
    """
    changes_old, changes_new = Hash(), Hash()

    old_paths = old.paths(intermediate=False)
    new_paths = new.paths(intermediate=False)

    keys = sorted(set(old_paths).union(set(new_paths)))
    for key in keys:
        old_value = old.get(key, None)
        new_value = new.get(key, None)
        if isinstance(old_value, (bytes, bytearray)):
            continue
        if old_value is None and new_value is not None:
            if not project:
                changes_old[key] = "Missing from configuration"
            changes_new[key] = new_value
        elif old_value is not None and new_value is None:
            changes_old[key] = old_value
            changes_new[key] = "Removed from configuration"
        elif has_changes(old_value, new_value):
            changes_old[key] = old_value
            changes_new[key] = new_value

    return changes_old, changes_new



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from qtpy.QtCore import QRectF
from qtpy.QtGui import QBrush, QColor, QLinearGradient, QPainter, QPen, QPixmap

from karabogui.graph.common.api import COLORMAPS
from karabogui.testing import GuiTestCase

from ..plot import HistogramPlot

LENGTH = 100
LEVELS = (0, 255)


class TestHistogramPlot(GuiTestCase):

    def setUp(self):
        super().setUp()
        self._plot_item = HistogramPlot(orientation="top")
        self._colormap = COLORMAPS[self._plot_item.colormap]

    def test_basics(self):
        np.testing.assert_array_equal(self._plot_item.levels, LEVELS)
        assert isinstance(self._plot_item.gradient, QLinearGradient)
        assert isinstance(self._plot_item._pen, QPen)
        assert self._plot_item.plotItem.vb.menu is None

    def test_set_colormap(self):
        cmap = "magma"
        self._plot_item.colormap = cmap
        stops = [(stop, color.getRgb()[:3])
                 for stop, color in self._plot_item.gradient.stops()]
        assert stops == COLORMAPS[cmap]

    def test_set_data(self):
        # 1. Set data equal to levels
        self._assert_gradient(start_value=0, stop_value=255)

        # 2. Set data inside levels range
        self._assert_gradient(start_value=50, stop_value=200)

        # 3. Set data outside levels range
        self._assert_gradient(start_value=-50, stop_value=300)

    def _assert_gradient(self, *, start_value, stop_value):
        x_data = np.linspace(start_value, stop_value, LENGTH)
        self._plot_item.set_data(x_data, np.arange(LENGTH - 1))

        # Get the index with respect to levels
        start_index = int(max(start_value, LEVELS[0]))
        stop_index = int(min(stop_value, LEVELS[1]))

        start_color, stop_color = self._get_border_colors()

        np.testing.assert_allclose(start_color, self._colormap[start_index][1],
                                   atol=4)
        np.testing.assert_allclose(stop_color, self._colormap[stop_index][1],
                                   atol=4)

    def _get_border_colors(self):
        """Get gradient start and final stop colors by painting its QPixmap
           and getting the color values from its QImage equivalent."""
        size = (1000, 10)
        brush = QBrush(self._plot_item.gradient)
        pixmap = QPixmap(*size)

        painter = QPainter(pixmap)
        painter.fillRect(QRectF(0, 0, *size), brush)
        painter.end()

        qimage = pixmap.toImage()
        start_color = QColor(qimage.pixel(1, 0)).getRgb()[:3]
        stop_color = QColor(qimage.pixel(999, 0)).getRgb()[:3]

        return start_color, stop_color



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.common.project.device import DeviceInstanceModel
from karabo.common.project.device_config import DeviceConfigurationModel
from karabo.common.project.server import DeviceServerModel
from karabo.native import Hash
from karabogui.project.controller.device import DeviceInstanceController
from karabogui.project.controller.server import DeviceServerController
from karabogui.singletons.mediator import Mediator
from karabogui.testing import singletons, system_hash
from karabogui.topology.system_topology import SystemTopology

from .utils import status_icon_path

PARENT_PATH = "karabogui.project.controller.device.find_parent_object"


def test_empty_fields_configurator(mocker):
    device_icon = mocker.patch(status_icon_path("device"))
    device_icon.return_value = None

    network = mocker.Mock()
    mediator = Mediator()
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(network=network, topology=topology, mediator=mediator):
        foo_config = Hash()
        foo_config["test1"] = "value1"
        foo_config["test2"] = "value2"
        foo_config = DeviceConfigurationModel(
            class_id="BazClass",
            configuration=foo_config)
        foo_config.initialized = True

        bar_config = Hash()
        bar_config["test1"] = "diff_value1"
        bar_config = DeviceConfigurationModel(
            class_id="BazClass",
            configuration=bar_config)
        bar_config.initialized = True

        config_list = [foo_config, bar_config]

        device_model = DeviceInstanceModel(
            class_id="BazClass",
            server_id="testServer",
            instance_id="fooDevice",
            configs=config_list)
        device_model.initialized = True
        device_model.active_config_ref = foo_config.uuid

        def assert_active_configuration():
            assert "test1" in foo_config.configuration
            assert foo_config.configuration["test1"] == "value1"

            assert "test2" in foo_config.configuration
            assert foo_config.configuration["test2"] == "value2"

            assert "test1" in bar_config.configuration
            assert bar_config.configuration["test1"] == "diff_value1"

            assert "test2" not in bar_config.configuration

        controller = DeviceInstanceController(model=device_model)
        controller.active_config_changed(bar_config)
        assert_active_configuration()

        controller.active_config_changed(foo_config)
        assert_active_configuration()

        controller.active_config_changed(bar_config)
        assert_active_configuration()


def test_move_devices(mocker):
    device_icon = mocker.patch(status_icon_path("device"))
    device_icon.return_value = None
    server_icon = mocker.patch(status_icon_path("device"))
    server_icon.return_value = None

    network = mocker.Mock()
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(network=network, topology=topology):
        # 1. Foo device with config
        foo_config = DeviceConfigurationModel(
            class_id="BazClass",
            configuration=Hash("value", "1"))
        foo_config.initialized = True
        foo_model = DeviceInstanceModel(
            class_id="BazClass",
            instance_id="fooDevice",
            server_id="testServer",
            configs=[foo_config])
        foo_model.initialized = True
        foo_model.active_config_ref = foo_config.uuid

        # 2. Bar device with config
        bar_config = DeviceConfigurationModel(
            class_id="BazClass",
            configuration=Hash("value", "2"))
        bar_config.initialized = True
        bar_model = DeviceInstanceModel(
            class_id="BazClass",
            server_id="testServer",
            instance_id="fooDevice2",
            configs=[bar_config])
        bar_model.initialized = True
        bar_model.active_config_ref = bar_config.uuid

        # 3. Server
        server_model = DeviceServerModel(
            server_id="testServer", host="serverFoo",
            devices=[foo_model, bar_model])
        server_model.initialized = True

        # 4. Controllers
        foo_controller = DeviceInstanceController(model=foo_model)
        bar_controller = DeviceInstanceController(model=bar_model)
        server_controller = DeviceServerController(model=server_model)

        assert server_controller.model.devices[0] == foo_model
        assert server_controller.model.devices[1] == bar_model

        # 5. Fake project class for testing with a model member
        class Project:
            model = None

        project = Project()

        find_parent = mocker.patch(PARENT_PATH)
        find_parent.return_value = server_model
        foo_controller._move_down(project)

        assert server_controller.model.devices[0] == bar_model
        assert server_controller.model.devices[1] == foo_model

        foo_controller._move_up(project)

        assert server_controller.model.devices[0] == foo_model
        assert server_controller.model.devices[1] == bar_model

        bar_controller._move_up(project)

        assert server_controller.model.devices[0] == bar_model
        assert server_controller.model.devices[1] == foo_model

        # Lets try to crack it and move out of range (< 0)
        bar_controller._move_up(project)

        assert server_controller.model.devices[0] == bar_model
        assert server_controller.model.devices[1] == foo_model

        # Lets try to crack it and move out of range (> len)
        foo_controller._move_down(project)

        assert server_controller.model.devices[0] == bar_model
        assert server_controller.model.devices[1] == foo_model



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from scipy.optimize import curve_fit
from traits.api import (
    Any, Array, ArrayOrNone, Bool, Callable, Dict, Property, cached_property)

from karabogui.graph.common.api import ImageRegion
from karabogui.graph.common.fitting import gaussian_fit as gauss1d

from ..base.analyzer import BaseAnalyzer

FWHM_COEFF = 2 * np.sqrt(2 * np.log(2))


class ProfileAnalyzer(BaseAnalyzer):
    # The stats depends on the fit parameters
    stats = Property(Dict, depends_on='_fit_params')

    # Smoothen the resulting signal
    smooth = Bool(False)

    # Transformed image axis data
    axis_data = ArrayOrNone

    # The profiling function
    _profile_function = Callable(np.average)

    # --- Arrays ---
    # Resulting arrays from `analyze`
    _x_profile = Array
    _y_profile = Array

    # Resulting arrays from `fit`, which depends on the profile arrays
    _x_fit = Array
    _y_fit = Array

    # Fit parameters from the gaussian fit. Either tuple of values or None.
    _fit_params = Any

    # --- Convenience checkers ---
    has_profile = Property(Bool)
    has_fit = Property(Bool)

    # -----------------------------------------------------------------------
    # Public methods

    def analyze(self, region, axis=0):
        """Calculate the profile of the region on each axes."""
        x_profile, y_profile = np.array([]), np.array([])

        # Check if there's existing data. Bail out if None.
        if self.axis_data is not None:
            axis_slice = region.slices[axis]

            # Obtain the y-axis profile, which is dependent on the
            # image region type
            if region.region_type is ImageRegion.Area:
                y_profile = self._profile_function(region.region, axis=axis)
                x_profile = self.axis_data[axis_slice]
            elif region.region_type is ImageRegion.Line:
                x_profile = self.axis_data
                y_profile = region.region[axis]

                if self.smooth and len(y_profile) > 200:
                    y_profile = smooth_signal(y_profile)

        # Calculate the profile with the profiling function (e.g., sum, mean)
        self._x_profile, self._y_profile = x_profile, y_profile
        return x_profile, y_profile

    def fit(self):
        """Use Gaussian function to fit the profile"""
        # Calculate the offset of the x-axis with the half of the difference
        # of data points (it is assumed that the data is uniformly spaced)
        x_fit, y_fit = np.array([]), np.array([])
        fit_params = None

        if self.has_profile:
            offset = abs(self._x_profile[1] - self._x_profile[0]) / 2
            fit_params = gaussian_fit(self._x_profile, self._y_profile,
                                      offset=offset)

            if fit_params is not None:
                x_fit = self.axis_data + offset
                y_fit = gauss1d(self.axis_data, *fit_params)

        self._x_fit, self._y_fit = x_fit, y_fit
        self._fit_params = fit_params
        return x_fit, y_fit

    def clear_data(self):
        """Clear all the calculated arrays and the fit parameters"""
        self._x_profile = np.array([])
        self._y_profile = np.array([])
        self._x_fit = np.array([])
        self._y_fit = np.array([])
        self._fit_params = None

    # -----------------------------------------------------------------------
    # Trait properties

    @cached_property
    def _get_stats(self):
        stats = {}

        if self.has_profile:
            # Select data to calculate statistics on
            x_data, y_data = self._x_profile, self._y_profile
            if self.has_fit:
                x_data, y_data = self._x_fit, self._y_fit

            index_max = np.argmax(y_data)

            stats["max_pos"] = x_data[index_max]
            stats["amplitude"] = y_data[index_max]

            if self.has_fit:
                stats["fwhm"] = FWHM_COEFF * self._fit_params[2]

        return stats

    def _get_has_fit(self):
        return self._fit_params is not None

    def _get_has_profile(self):
        return bool(self._x_profile.size and self._y_profile.size)


def smooth_signal(y_profile):
    """We smooth the received data using the Moving Average approach so
    we can get rid of the up/down peaks on noisy images that make the
    plot very time-consuming. Ideally this should be done only on noisy
    images, so once we figure out a way to assert it we can avoid doing
    this everytime"""
    window = int(0.005 * len(y_profile))  # Window size of 0.5%
    window = max(3, window)
    y_profile = np.convolve(y_profile,
                            np.ones((window,)) / window,
                            mode='same')

    return y_profile


def gaussian_fit(x_data, y_data, offset=0):
    """
    Centre-of-mass and width. Lifted from image_processing.imageCentreofMass()
    """
    try:
        x0 = np.average(x_data, weights=y_data)
        sx = np.sqrt(np.average((x_data - x0) ** 2, weights=y_data))
    except ZeroDivisionError:
        # Weighting might fail with zero's
        return None

    # Gaussian fit
    p_0 = (y_data.max(), x0 + offset, sx, y_data[0])
    try:
        p_f, _ = curve_fit(gauss1d, x_data, y_data, p_0)
        return p_f
    except (RuntimeError, TypeError):
        return None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import Qt, Slot
from qtpy.QtGui import QFont, QFontDatabase, QPalette, QValidator
from qtpy.QtWidgets import QDialog

from karabo.common.scenemodel.const import SCENE_FONT_SIZES
from karabogui.fonts import (
    FONT_ALIAS, get_alias_from_font, get_font_from_alias,
    get_font_size_from_dpi)
from karabogui.validators import IntValidator

from .utils import get_dialog_ui

DEFAULT_STYLE = "Regular"


class FontDialog(QDialog):

    def __init__(self, qfont, parent=None):
        super().__init__(parent=parent)
        uic.loadUi(get_dialog_ui("font_dialog.ui"), self)

        # Instantiate variables
        self.qfont = QFont(qfont)
        self._font_database = QFontDatabase()

        # Populate font family
        self.font_combobox.addItems(FONT_ALIAS)
        self.font_combobox.setCurrentText(get_alias_from_font(qfont.family()))

        # Populate font size combobox and set current index/text
        sizes_string = [str(size) for size in SCENE_FONT_SIZES]
        font_size_cb = self.font_size_combobox
        font_size_cb.addItems(sizes_string)
        font_size = qfont.pointSize()
        font_size_cb.setCurrentText(str(font_size))

        self._font_validator = IntValidator(minInc=6, maxInc=300)
        self._normal_palette = self.font_size_combobox.palette()
        self._error_palette = QPalette(self._normal_palette)
        self._error_palette.setColor(QPalette.Text, Qt.red)
        # attach validator for basic string protection
        self.font_size_combobox.setValidator(self._font_validator)

        # Update effects
        self.bold_checkbox.setChecked(qfont.bold())
        self.italic_checkbox.setChecked(qfont.italic())
        self.strikeout_checkbox.setChecked(qfont.strikeOut())
        self.underline_checkbox.setChecked(qfont.underline())

        # Add triggers to change preview
        self.font_combobox.currentTextChanged.connect(self._update_font)
        font_size_cb.currentTextChanged.connect(self._update_font_size)
        self.bold_checkbox.toggled.connect(self._update_bold)
        self.italic_checkbox.toggled.connect(self._update_italic)
        self.strikeout_checkbox.toggled.connect(self._update_strikeout)
        self.underline_checkbox.toggled.connect(self._update_underline)

        # Update preview to reflect current format
        self._preview()

    @Slot()
    def _preview(self):
        qfont = QFont(self.qfont)
        qfont.setPointSize(get_font_size_from_dpi(qfont.pointSize()))
        self.preview_label.setFont(qfont)

    @Slot(str)
    def _update_font(self, family):
        self.qfont.setFamily(get_font_from_alias(family))
        self._preview()

    @Slot(str)
    def _update_font_size(self, size):
        out, _, _ = self._font_validator.validate(size, None)
        validated = out is QValidator.Acceptable
        palette = (self._normal_palette if validated else self._error_palette)
        self.font_size_combobox.setPalette(palette)
        if validated:
            self.qfont.setPointSize(int(size))
            self._preview()

    @Slot(str)
    def _update_font_style(self, style):
        self.qfont.setStyleName(style)
        self._preview()

    @Slot(bool)
    def _update_bold(self, enabled):
        self.qfont.setBold(enabled)
        self._preview()

    @Slot(bool)
    def _update_italic(self, enabled):
        self.qfont.setItalic(enabled)
        self._preview()

    @Slot(bool)
    def _update_strikeout(self, enabled):
        self.qfont.setStrikeOut(enabled)
        self._preview()

    @Slot(bool)
    def _update_underline(self, enabled):
        self.qfont.setUnderline(enabled)
        self._preview()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on November 29, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy import uic
from qtpy.QtCore import Slot
from qtpy.QtGui import QValidator
from qtpy.QtWidgets import QDialog, QDialogButtonBox, QFormLayout, QSpinBox

from karabogui.util import InputValidator, SignalBlocker

from .utils import get_dialog_ui


class ObjectDuplicateDialog(QDialog):
    def __init__(self, simple_name, parent=None):
        """ The dialog expects a ``simple_name`` string

        :param simple_name: A string which could be used for creating the new
                            duplicate simple Name
        :param parent: A parent object
        """
        super().__init__(parent)
        uic.loadUi(get_dialog_ui('object_duplicate.ui'), self)
        self.setWindowTitle(f'Duplicate object {simple_name}')

        # Add spinboxes with validator
        self.sbStart = UIntSpinbox(self)
        self.sbStart.valueChanged.connect(self._indexChanged)
        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.sbStart)
        self.sbEnd = UIntSpinbox(self)
        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.sbEnd)
        self.sbEnd.valueChanged.connect(self._indexChanged)
        self.cbNoIndex.stateChanged.connect(self._updateIndex)

        validator = InputValidator(parent=self)
        self.leTitle.setValidator(validator)
        self.leTitle.setText(simple_name)
        self._update_text()

    def _update_text(self):
        """Update the text for duplicated devices"""
        if not self.cbNoIndex.isChecked():
            start_index = self.sbStart.value()
            end_index = self.sbEnd.value()
            nb_dupe = end_index - start_index + 1
        else:
            nb_dupe = 1

        text = f'You are about to create <b>{nb_dupe}</b> duplicate(s)'
        self.laText.setText(text)

    @Slot(int)
    def _updateIndex(self, value):
        enabled = True if not value else False
        self.sbStart.setEnabled(enabled)
        self.sbEnd.setEnabled(enabled)
        self._update_text()

    @Slot(int)
    def _indexChanged(self, value):
        start_index = self.sbStart.value()
        end_index = self.sbEnd.value()

        while start_index > end_index:
            end_index += 1
        with SignalBlocker(self.sbEnd):
            self.sbEnd.setValue(end_index)
        self._update_text()

    @property
    def duplicate_names(self):
        """Create duplicated names for the project devices

        If the `noIndex` checkbox is enabled, we only consider the `title`!
        """
        simple_names = []
        if not self.cbNoIndex.isChecked():
            start_index = self.sbStart.value()
            end_index = self.sbEnd.value()
            for index in range(start_index, end_index + 1):
                dupe_name = f'{self.leTitle.text()}{index}'
                simple_names.append(dupe_name)
        else:
            simple_names.append(self.leTitle.text())

        return simple_names


class UIntSpinbox(QSpinBox):

    def __init__(self, parent):
        super().__init__(parent)
        self.setMinimum(0)
        self.setMaximum(1000)

    def validate(self, value, pos):
        if not value.isdigit() and value != '':
            return QValidator.Invalid, value, pos

        return super().validate(value, pos)


class ObjectEditDialog(QDialog):
    def __init__(self, object_type='object', model=None, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui('object_edit.ui'), self)

        if model is None:
            title = f'Add {object_type}'
        else:
            title = f'Edit {object_type}'
            self.leTitle.setText(model.simple_name)

        validator = InputValidator(object_type, parent=self)
        self.leTitle.setValidator(validator)
        self.leTitle.textChanged.connect(self.validate)
        self.setWindowTitle(title)
        self.validate()

    @Slot()
    def validate(self):
        enabled = self.leTitle.hasAcceptableInput()
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)

    @property
    def simple_name(self):
        return self.leTitle.text()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import QItemSelectionModel

from karabogui.singletons.selection_tracker import SelectionTracker


def test_selection_tracker(mocker):
    """Test the Selection Tracker singleton"""
    tracker = SelectionTracker()

    model = mocker.Mock()
    # Must be a `QItemSelectionModel`
    with pytest.raises(AssertionError):
        tracker.grab_selection(model)
    assert tracker._model is None

    model = QItemSelectionModel()
    model.clear = mocker.Mock()
    tracker.grab_selection(model)
    assert tracker._model is model

    model.clear.assert_not_called()
    new_model = QItemSelectionModel()
    new_model.clear = mocker.Mock()
    tracker.grab_selection(new_model)
    model.clear.assert_called_once()
    # Grabbing the selection only resets once the old model
    tracker.grab_selection(new_model)
    model.clear.assert_called_once()

    # New model is not reset!
    new_model.clear.assert_not_called()
    assert tracker._model is new_model


if __name__ == "__main__":
    pytest.main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import QItemSelectionModel
from qtpy.QtWidgets import QApplication

from karabo.native import Hash
from karabogui.events import KaraboEvent
from karabogui.testing import GuiTestCase
from karabogui.widgets.log import LogWidget


def create_log_data():
    logs = []

    for i in range(100):
        gui_server = Hash(
            "timestamp", "2021-07-05T15:08:35.250",
            "type", "INFO",
            "category", f"Karabo_GuiServer_{i}",
            "message", "Hello!")
        logs.append(gui_server)

        debug_log = Hash(
            "timestamp", "2021-07-05T13:09:25.120",
            "type", "DEBUG",
            "category", f"XFEL_EG_RACK/MOTOR/{i}",
            "message", "Schema Updated")
        logs.append(debug_log)

        error_log = Hash(
            "timestamp", "2021-07-05T13:09:25.120",
            "type", "ERROR",
            "category", f"XFEL_EG_RACK/MOTOR/{i}",
            "message", "Error in background task",
            "traceback", "This is a traceback"
        )
        logs.append(error_log)

    return logs


class TestLogWidget(GuiTestCase):

    def test_log_widget_prune(self):
        widget = LogWidget()
        model = widget.table.model()
        data = create_log_data()
        widget.onLogDataAvailable(data)
        assert model.rowCount() == 300
        # add 300 more to overshoot and prune, stays at 300
        widget.onLogDataAvailable(data)
        assert model.rowCount() == 300
        assert model.rowCount() > 0
        widget.onClearLog()
        assert model.rowCount() == 0
        # Do a reset with new data
        widget.initialize(data)
        assert model.rowCount() == 300

    def test_clipboard_doubleclick(self):
        widget = LogWidget()
        data = create_log_data()
        widget.onLogDataAvailable(data)
        path = "karabogui.widgets.log.broadcast_event"
        with mock.patch(path) as broadcast:
            source_index = widget.table_model.createIndex(12, 0)
            index = widget.filter_model.mapFromSource(source_index)
            widget.onItemDoubleClicked(index)
            broadcast.assert_called_with(KaraboEvent.ShowDevice,
                                         {"deviceId": "XFEL_EG_RACK/MOTOR/95"})

        selection_model = widget.table.selectionModel()
        selection_model.setCurrentIndex(index,
                                        QItemSelectionModel.ClearAndSelect)

        widget._copy_clipboard()
        clipboard = QApplication.clipboard()
        text = clipboard.text()
        assert "XFEL_EG_RACK/MOTOR/95" in text


if __name__ == "__main__":
    main()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on October 6, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from enum import Enum
from pathlib import Path

from qtpy.QtGui import QIcon

import karabogui.access as krb_access
from karabo.common.api import InstanceStatus, State
from karabogui.binding.api import ProxyStatus

from . import icons

# --------------------------------------------------------------------------
# Mapping states to colors


def get_state_color(value):
    """Return a state color for a given state string

    :param value: state string representation
    :type value: str
    """
    state = State(value)
    if state.isDerivedFrom(State.CHANGING):
        color = STATE_COLORS[State.CHANGING]
    elif state.isDerivedFrom(State.RUNNING):
        color = STATE_COLORS[State.RUNNING]
    elif state.isDerivedFrom(State.ACTIVE):
        color = STATE_COLORS[State.ACTIVE]
    elif state.isDerivedFrom(State.PASSIVE):
        color = STATE_COLORS[State.PASSIVE]
    elif state.isDerivedFrom(State.DISABLED):
        color = STATE_COLORS[State.DISABLED]
    elif state in [State.STATIC, State.NORMAL, State.ERROR, State.INIT]:
        color = STATE_COLORS[state]
    else:
        color = STATE_COLORS[State.UNKNOWN]

    return color


class _StateColors(Enum):
    UNKNOWN_COLOR = (255, 170, 0)
    KNOWN_NORMAL_COLOR = (200, 200, 200)
    INIT_COLOR = (230, 230, 170)
    DISABLED_COLOR = (255, 0, 255)
    ERROR_COLOR = (255, 0, 0)
    CHANGING_DECREASING_INCREASING_COLOR = (0, 170, 255)
    RUNNING_COLOR = (153, 204, 255)
    STATIC_COLOR = (0, 170, 0)
    ACTIVE_COLOR = (120, 255, 0)
    PASSIVE_COLOR = (204, 204, 255)


STATE_COLORS = {
    State.UNKNOWN: _StateColors.UNKNOWN_COLOR.value,
    State.KNOWN: _StateColors.KNOWN_NORMAL_COLOR.value,
    State.NORMAL: _StateColors.KNOWN_NORMAL_COLOR.value,
    State.INIT: _StateColors.INIT_COLOR.value,
    State.DISABLED: _StateColors.DISABLED_COLOR.value,
    State.ERROR: _StateColors.ERROR_COLOR.value,
    State.CHANGING: _StateColors.CHANGING_DECREASING_INCREASING_COLOR.value,
    State.DECREASING: _StateColors.CHANGING_DECREASING_INCREASING_COLOR.value,
    State.INCREASING: _StateColors.CHANGING_DECREASING_INCREASING_COLOR.value,
    State.RUNNING: _StateColors.RUNNING_COLOR.value,
    State.STATIC: _StateColors.STATIC_COLOR.value,
    State.ACTIVE: _StateColors.ACTIVE_COLOR.value,
    State.PASSIVE: _StateColors.PASSIVE_COLOR.value
}


# --------------------------------------------------------------------------
# Mapping device status to icons


def get_instance_info_icon(status):
    """Return an instance info icon which reflects the given `DeviceStatus`
    """
    if not isinstance(status, InstanceStatus):
        status = InstanceStatus(status)

    if status is InstanceStatus.OK:
        return icons.statusOk
    elif status is InstanceStatus.ERROR:
        return icons.statusError
    elif status is InstanceStatus.UNKNOWN:
        return icons.statusUnknown
    # No icon!
    return None


def get_device_status_icon(status):
    """A `QIcon` for the given `status` is returned."""
    status_icons = {
        ProxyStatus.OFFLINE: icons.deviceOffline,
        ProxyStatus.ALIVE: icons.deviceAlive,
        ProxyStatus.REQUESTED: icons.device_requested,
        ProxyStatus.ONLINEREQUESTED: icons.device_requested,
        ProxyStatus.SCHEMA: icons.device_schema,
        ProxyStatus.NOSERVER: icons.deviceOfflineNoServer,
        ProxyStatus.NOPLUGIN: icons.deviceOfflineNoPlugin,
        ProxyStatus.INCOMPATIBLE: icons.deviceIncompatible,
        ProxyStatus.MISSING: icons.propertyMissing,
    }

    if not isinstance(status, ProxyStatus):
        status = ProxyStatus(status)

    return status_icons.get(status)


def get_device_status_pixmap(status, extent=16):
    """A `QPixmap` for the given `status` is returned.

    `extent` sets the size of the pixmap. The pixmap might be smaller than
    requested, but never larger.
    """
    icon = get_device_status_icon(status)
    if icon is not None:
        return icon.pixmap(extent)


def get_project_device_status_icon(status, error=False):
    """An icon for the given device `status`.

    NOTE: These icons are for the Project Panel
    """
    status_icons = {
        ProxyStatus.NOSERVER: icons.deviceOfflineNoServer,
        ProxyStatus.NOPLUGIN: icons.deviceOfflineNoPlugin,
        ProxyStatus.OFFLINE: icons.deviceOffline,
        ProxyStatus.REQUESTED: icons.deviceOffline,
        ProxyStatus.INCOMPATIBLE: icons.deviceIncompatible,
        ProxyStatus.MONITORING: icons.deviceMonitored,
    }

    if not isinstance(status, ProxyStatus):
        status = ProxyStatus(status)

    return status_icons.get(status, icons.deviceInstance)


def get_project_server_status_icon(status):
    """An icon for the given server `status`.

    NOTE: These icons are for the Project Panel
    """
    status_icons = {
        ProxyStatus.ONLINE: icons.yes,
        ProxyStatus.OFFLINE: icons.no,
    }

    if not isinstance(status, ProxyStatus):
        status = ProxyStatus(status)

    return status_icons.get(status)


# ---------------------------------
# Lamp coloring in the main window

PROC_FINE = 2
PROC_ALARM = 5
PROC_FINE_COLOR = (120, 255, 0, 200)
PROC_BETWEEN_COLOR = (255, 255, 102, 200)
PROC_ALARM_COLOR = (255, 0, 0, 200)


def get_processing_color(proc_delay):
    """Define the processing lamp coloring in the main window

    A three stage lamp is provided with basic colors: green, yellow, red
    """
    if proc_delay < PROC_FINE:
        return PROC_FINE_COLOR
    elif PROC_FINE <= proc_delay <= PROC_ALARM:
        return PROC_BETWEEN_COLOR
    elif proc_delay > PROC_ALARM:
        return PROC_ALARM_COLOR


# --------------------------------------------------------------------------
# Mapping topics to colors

def get_topic_color(topic=None):
    """Get the corresponding standard color belonging to karabo topic

    :param topic: The karabo topic string, e.g. SA1, FXE, SPB, SA2, etc.
    """
    color_map = {
        "SA1": (140, 170, 215, 128),
        "SA2": (215, 170, 150, 128),
        "SA3": (150, 185, 150, 128),
    }

    return color_map.get(topic, None)


# --------------------------------------------------------------------------
# Commonly used colors

OK_COLOR = (225, 242, 225, 128)
PROPERTY_READONLY_COLOR = (160, 160, 160, 255)
ALL_OK_COLOR = (224, 224, 224, 128)  # no alarm and fine
ERROR_COLOR_ALPHA = (255, 155, 155, 128)  # semitransparent
UNKNOWN_COLOR_ALPHA = (255, 200, 150, 128)
PROPERTY_ALARM_COLOR = (255, 125, 125, 128)
PROPERTY_WARN_COLOR = (255, 255, 125, 128)
PROPERTY_INTERLOCK_COLOR = (51, 51, 255, 128)
LOCKED_COLOR = (255, 145, 255, 128)

# --------------------------------------------------------------------------
# Mapping alarms to colors

WARN_GLOBAL = "warn"
ALARM_GLOBAL = "alarm"
INTERLOCK = "interlock"
ALARM_NONE = "none"

PROPERTY_ALARM_COLOR_MAP = {
    None: ALL_OK_COLOR,
    ALARM_NONE: None,
    ALARM_GLOBAL: PROPERTY_ALARM_COLOR,
    WARN_GLOBAL: PROPERTY_WARN_COLOR,
    INTERLOCK: PROPERTY_INTERLOCK_COLOR}

ICON_PATH = Path(icons.__file__).parent


def get_alarm_svg(alarm_type):
    """The svg icon for the given `alarm_type` is returned.
    """
    svg_none = str(ICON_PATH.joinpath("alarm_none.svg"))
    svg_warn = str(ICON_PATH.joinpath("warning.svg"))
    svg_alarm = str(ICON_PATH.joinpath("critical.svg"))
    svg_interlock = str(ICON_PATH.joinpath("interlock.svg"))

    ALARM_SVG = {
        ALARM_NONE: svg_none,
        WARN_GLOBAL: svg_warn,
        ALARM_GLOBAL: svg_alarm,
        INTERLOCK: svg_interlock,
    }

    alarm_svg = ALARM_SVG.get(alarm_type)
    if alarm_svg is not None:
        return alarm_svg


def get_user_session_button_data() -> (QIcon, str):
    """Provide the icon and tooltip for the temporary session button
    depending on the state of the temporary session."""
    if krb_access.SESSION_END_NOTICE:
        tooltip = "Need to reauthenticate."
        return icons.switchCritical, tooltip

    user = krb_access.TEMPORARY_SESSION_USER
    level = krb_access.GLOBAL_ACCESS_LEVEL.name
    if user is None:
        tooltip = "User Session Info"
        return icons.switchNormal, tooltip
    if krb_access.TEMPORARY_SESSION_WARNING:
        tooltip = (f"Logged in as '{user}' with access level '{level}'. "
                   "Temporary session is about to end.")
        return icons.switchWarning, tooltip
    tooltip = (f"Logged in as '{user}' with access level '{level}'. "
               "End the temporary session.")
    return icons.switchActive, tooltip



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from collections import defaultdict

from pyqtgraph import Point
from qtpy.QtCore import QObject, Signal, Slot

from karabogui.graph.common.enums import ROITool
from karabogui.graph.common.roi.crosshair import CrosshairROI
from karabogui.graph.common.roi.rect import RectROI


class BaseROIController(QObject):
    """
    This class is responsible for managing all the ROIs of a plotItem.
    Therefore, all ROIs should be added/removed from here. It offers 4
    signals for communicating to the outside:
        - updated: when the ROI information is refreshed
        - removed: when an ROI is removed
        - selected: when one of the ROIs is selected
    """
    updated = Signal(object)
    removed = Signal()
    selected = Signal(object)

    TOOL_MAP = {
        ROITool.Rect: RectROI,
        ROITool.Crosshair: CrosshairROI
    }

    def __init__(self, plotItem):
        super().__init__()
        self._rois = defaultdict(list)

        self.plotItem = plotItem

        self._current_tool = ROITool.NoROI
        self._current_item = defaultdict(lambda: None)

        self._updates_enabled = False

        # Position snapping is disabled by default.
        # This is usually utilized by image ROIs
        self.scaleSnap = False
        self.translateSnap = False

        self._pen = None

    # ---------------------------------------------------------------------
    # Public methods

    @property
    def roi_items(self):
        return dict(self._rois)

    @property
    def current_tool(self):
        return self._current_tool

    @property
    def current_tool_item(self):
        return self._current_item[self._current_tool]

    def show(self, roi):
        """
        Enables the given ROITool.
        """
        self._current_tool = roi
        if roi == ROITool.NoROI:
            self.hide_all()
        else:
            for class_ in self._rois.keys():
                if roi != class_:
                    self._hide_roi_class(class_)

            self._show_roi_class(roi)

    def add(self, tool, pos, size=None, name='', ignore_bounds=False,
            current_item=True):
        # Get the ROI class
        roi_class = self.TOOL_MAP[tool]
        pos = Point(pos)

        if size is not None:
            roi_item = roi_class(pos=pos, size=size, name=name,
                                 scaleSnap=self.scaleSnap,
                                 translateSnap=self.translateSnap,
                                 pen=self._pen)
        else:
            roi_item = roi_class(pos=pos, name=name,
                                 scaleSnap=self.scaleSnap,
                                 translateSnap=self.translateSnap,
                                 pen=self._pen)

        # Connect some signals
        roi_item.sigRegionChangeStarted.connect(self._set_current_item)
        roi_item.sigRegionChanged.connect(self.update)
        roi_item.sigRemoveRequested.connect(self._remove_roi_item)
        roi_item.sigClicked.connect(self._set_current_item)

        # Bookkeeping. Add the items to the list of existing ROI tools
        self._rois[tool].append(roi_item)

        # Add the ROI item to the plot
        self._add_to_plot(roi_item, ignore_bounds)

        if current_item:
            # Set as current item, in which affects the aux plots.
            self._current_tool = tool
            self._set_current_item(roi_item, update=False)

        return roi_item

    def set_pen(self, pen):
        self._pen = pen

    def _add_to_plot(self, roi_item, ignore_bounds=True):
        self.plotItem.vb.addItem(roi_item, ignoreBounds=ignore_bounds)
        if roi_item.textItem:
            self.plotItem.vb.addItem(roi_item.textItem,
                                     ignoreBounds=ignore_bounds)

    def _remove_from_plot(self, roi_item):
        roi_item.sigRegionChanged.disconnect()
        roi_item.sigRegionChangeStarted.disconnect()
        roi_item.sigClicked.disconnect()
        roi_item.set_visible(False)

        if roi_item.textItem:
            self.plotItem.vb.removeItem(roi_item.textItem)
            roi_item.textItem.deleteLater()

        self.plotItem.vb.removeItem(roi_item)
        roi_item.deleteLater()

    def hide_all(self):
        """
        Hide all added ROI classes.
        """
        for roi_class in self._rois.keys():
            self._hide_roi_class(roi_class)

    def remove_all(self):
        for roi_items in self._rois.values():
            for item in roi_items:
                self._remove_roi_item(item)

        self._rois.clear()

    def get_coords(self):
        """Retrieves the absolute coords of all existing ROIs and returns dict:
           {ROITool.Crosshair: [(0,0), (1,1), ..], ROITool.Rect: [(0,0,0,0)]}
        """
        roi_coords = {}
        for roi, items in self._rois.items():
            roi_coords[roi] = [item.coords for item in items]
        return roi_coords

    def destroy(self):
        self.updated.disconnect()
        self.deleteLater()

    def enable_updates(self, enable=True):
        self._updates_enabled = enable
        self.update()

    @Slot()
    def update(self):
        """Subclass for ROI region change interactions"""

    # ---------------------------------------------------------------------
    # Private methods

    def _hide_roi_class(self, roi_class):
        """
        Hides the ROI belonging to the given class.
        """
        roi_items = self._rois[roi_class]
        if not roi_items:
            return

        for item in roi_items:
            item.set_visible(False)

    @Slot(object)
    def _remove_roi_item(self, roi_item):
        """
        Deletes the given ROI and removes it from the list.
        """
        # Delete the item from the list
        roi_items = self._rois[self._current_tool]
        roi_items.remove(roi_item)

        # Check first if current item
        if roi_item is self._current_item[self._current_tool]:
            # Give the honor to others if available.
            if roi_items:
                self._set_current_item(roi_items[-1])
            else:
                self._set_current_item(None)
                self.removed.emit()

        # Then delete the item
        self._remove_from_plot(roi_item)

    def _show_roi_class(self, roi_class):
        """
        Show a ROI item for a given class, creating a new one if it's not
        available.
        """
        # Use already created ROI if available
        roi_items = self._rois[roi_class]

        # Do nothing if no recorded items
        if not roi_items:
            return

        for item in roi_items:
            self._show_roi_item(item)

        # Set last item as current if None
        if self._current_item[roi_class] is None:
            self._set_current_item(item)

    def _show_roi_item(self, roi_item):
        """ Shows an ROI item """
        roi_item.set_visible(True)

    @Slot(object)
    def _set_current_item(self, roi_item, update=True):
        # Check if item is the same as current
        if roi_item is self._current_item[self._current_tool]:
            return

        # Deselect current ROI
        current_item = self._current_item[self._current_tool]
        if current_item is not None:
            current_item.selected = False

        # Put new ROI as selected
        self._current_item[self._current_tool] = roi_item
        if roi_item is None:
            self._current_tool = ROITool.NoROI
        else:
            roi_item.selected = True



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtGui import QColor

from karabogui.graph.plots.api import GraphViewDialog
from karabogui.testing import GuiTestCase


class TestDialogs(GuiTestCase):

    def test_view_dialog(self):
        # Test no config dialog
        dialog = GraphViewDialog()
        assert dialog.graph_bg_color == "transparent"
        assert not dialog.ui_cb_background.isChecked()

        settings = dialog.settings
        assert settings["background"] == "transparent"
        assert settings["title"] == ""

        path = "karabogui.graph.plots.dialogs.view.QColorDialog"
        with mock.patch(path) as color:
            c = QColor(0, 0, 255)
            color.getColor.return_value = c

            # Enable clicking dialog background
            self.click(dialog.ui_cb_background)
            self.click(dialog.ui_pb_background)
            self.process_qt_events()
            assert dialog.graph_bg_color == c.name()

            settings = dialog.settings
            assert settings["background"] == c.name()
            assert settings["title"] == ""

            # Disable background again
            self.click(dialog.ui_cb_background)
            settings = dialog.settings
            assert settings["background"] == "transparent"
            assert settings["title"] == ""

            title = "XRAY"
            dialog.ui_title.setText(title)
            settings = dialog.settings
            assert settings["background"] == "transparent"
            assert settings["title"] == title


if __name__ == "__main__":
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os

from qtpy import uic
from qtpy.QtWidgets import QDialog


class AxesLabelsDialog(QDialog):
    """This class dialog is responsible to provide editing of the axes labels
    """
    def __init__(self, config={}, parent=None):
        super().__init__(parent)
        self.setModal(False)
        # load ui file
        ui_path = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                               'axes_labels.ui')
        uic.loadUi(ui_path, self)

        self.ui_xlabel.setText(config.get('x_label', ''))
        self.ui_ylabel.setText(config.get('y_label', ''))
        self.ui_xunits.setText(config.get('x_units', ''))
        self.ui_yunits.setText(config.get('y_units', ''))

        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

    @property
    def labels(self):
        config = {
            "x_label": self.ui_xlabel.text(),
            "x_units": self.ui_xunits.text(),
            "y_label": self.ui_ylabel.text(),
            "y_units": self.ui_yunits.text()}

        return config

    @staticmethod
    def get(configuration, parent=None):
        dialog = AxesLabelsDialog(configuration, parent)
        result = dialog.exec() == QDialog.Accepted
        content = {}
        content.update(dialog.labels)

        return content, result



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from qtpy.QtGui import QImage

from karabogui.graph.image.api import karabo_default_image
from karabogui.graph.image.base import KaraboImageView
from karabogui.testing import GuiTestCase

# image dimensions
SMALL = (150, 100)
MEDIUM = (800, 600)
LARGE = (1600, 1200)
DIMENSIONS = [SMALL, MEDIUM, LARGE]
MIN_DOWNSAMPLING = [1, 1.5, 2]


class _BaseImageItemTest(GuiTestCase):
    def setUp(self):
        super().setUp()
        # Instantiate imageItem from KaraboImageView
        self.imageView = KaraboImageView()
        self.imageItem = self.imageView.plot().imageItem
        self.imageView.setFixedSize(300, 300)

    def tearDown(self):
        super().tearDown()
        self.imageView.destroy()
        self.imageView = None

    def set_image(self, image):
        # Set image data
        self.imageItem.setImage(image)
        self.process_qt_events()
        np.testing.assert_array_equal(image, self.imageItem.image)
        assert self.imageItem.qimage is None

        # Render QImage
        self.imageItem._viewBox().autoRange()
        self.imageItem._view_rect = self.imageItem.viewRect()
        self.imageItem.render()

    def assert_indexed_image(self, image, downsampling):
        self.set_image(image)
        self.assert_downsample(downsampling)
        self.assert_qimage(image, img_format=QImage.Format_Indexed8)

    def assert_rgb_image(self, image):
        self.set_image(image)
        self.assert_downsample(downsampling=None)
        self.assert_qimage(image, img_format=QImage.Format_ARGB32)

    def assert_downsample(self, downsampling):
        """Check downsample ratio and the resulting qimage dimensions"""
        if downsampling is None:
            assert self.imageItem._lastDownsample is None
        else:
            xds, yds = self.imageItem._lastDownsample
            assert xds >= downsampling
            assert yds >= downsampling

    def assert_qimage(self, image, *, img_format):
        qimage = self.imageItem.qimage
        height, width = image.shape[:2]
        if img_format == QImage.Format_Indexed8:
            # We expect that indexed format is downsampled
            xds, yds = self.imageItem._lastDownsample
            width = round(width / xds)
            height = round(height / yds)

        assert qimage.width() == width
        assert qimage.height() == height


class TestKaraboImageItem(_BaseImageItemTest):

    def test_rgb_images(self):
        for dims in DIMENSIONS:
            image = np.random.randint(0, 256, dims + (3,), dtype=np.uint8)
            self.assert_rgb_image(image)

    def test_rgba_images(self):
        for dims in DIMENSIONS:
            image = np.random.randint(0, 256, dims + (4,), dtype=np.uint8)
            self.assert_rgb_image(image)

    def test_alternating_image_types(self):
        """This checks when the image item receives alternating pseudocolor
           (indexed) and RGB images"""

        indexed_image = np.random.randint(0, 256, MEDIUM, dtype=np.uint8)
        rgb_image = np.random.randint(0, 256, MEDIUM + (4,), dtype=np.uint8)

        self.assert_indexed_image(indexed_image, downsampling=1.5)
        self.assert_rgb_image(rgb_image)
        self.assert_indexed_image(indexed_image, downsampling=1.5)

    def test_translation(self):
        """Test the translation of an image which is used by external clients
        """
        received = False

        def slotTransform():
            nonlocal received
            received = True

        dims = (800, 600)
        image = np.random.randint(0, 256, dims + (3,), dtype=np.uint8)
        self.assert_rgb_image(image)

        plotItem = self.imageView.plot()
        plotItem.imageTransformed.connect(slotTransform)

        np.testing.assert_array_equal(self.imageItem._origin,
                                      np.array([0, 0]))
        plotItem.set_translation(x_translate=10, y_translate=20)
        self.process_qt_events()

        np.testing.assert_array_equal(self.imageItem._origin,
                                      np.array([10, 20]))
        assert received
        self.assert_rgb_image(image)
        plotItem.imageTransformed.disconnect(slotTransform)


class Test2DImageItem(_BaseImageItemTest):

    def test_uint8_images(self):
        for dims, downsampling in zip(DIMENSIONS, MIN_DOWNSAMPLING):
            image = np.random.randint(0, 255, dims, dtype=np.uint8)
            self.assert_indexed_image(image, downsampling)

    def test_unint32_images(self):
        for dims, downsampling in zip(DIMENSIONS, MIN_DOWNSAMPLING):
            image = np.random.randint(0, 100000, dims, dtype=np.uint32)
            self.assert_indexed_image(image, downsampling)

    def test_int8_images(self):
        for dims, downsampling in zip(DIMENSIONS, MIN_DOWNSAMPLING):
            image = np.random.randint(-128, 128, dims, dtype=np.int8)
            self.assert_indexed_image(image, downsampling)

    def test_int32_images(self):
        for dims, downsampling in zip(DIMENSIONS, MIN_DOWNSAMPLING):
            image = np.random.randint(-50000, 50000, dims, dtype=np.int32)
            self.assert_indexed_image(image, downsampling)

    def test_float_images(self):
        for dims, downsampling in zip(DIMENSIONS, MIN_DOWNSAMPLING):
            image = np.random.uniform(-50000, 50000, dims)
            self.assert_indexed_image(image, downsampling)

    def test_set_graph_image(self):
        """Make sure setting levels with default image (all zero array)"""
        image = karabo_default_image()
        self.imageItem.setImage(image)
        self.imageItem._set_graph_image(autoLevels=True)
        assert np.array_equal(self.imageItem.levels, np.array([0, 1]))



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QDialog

from karabogui.graph.common.api import AxesLabelsDialog


def test_view_dialog(gui_app, mocker):
    # Test no config dialog
    empty_config = {"x_label": "",
                    "y_label": "",
                    "x_units": "",
                    "y_units": ""}

    dialog = AxesLabelsDialog()
    assert dialog.windowTitle() == "Edit Axes Details"
    assert dialog.labels == empty_config

    config = {"x_label": "Time",
              "y_label": "#",
              "x_units": "s",
              "y_units": ""}
    dialog = AxesLabelsDialog(config=config)
    assert dialog.labels == config

    path = "karabogui.graph.common.dialogs.axes_labels.AxesLabelsDialog"
    d = mocker.patch(path)
    d().exec.return_value = QDialog.Accepted
    d().labels = dialog.labels
    # We fake a success return and test the boolean, content reply
    content, success = dialog.get(None, None)
    assert success
    assert content == config



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest.mock import patch

import numpy as np
from traits.api import Undefined

from karabo.common.api import State
from karabo.native import (
    AccessLevel, AccessMode, Assignment, Bool, Configurable, Float, Hash,
    HashList, String, UInt32, VectorBool, VectorHash, VectorUInt32)
from karabogui.binding.api import (
    DeviceClassProxy, DeviceProxy, ProjectDeviceProxy, ProxyStatus,
    VectorHashBinding, apply_configuration, apply_default_configuration,
    build_binding, validate_table_value, validate_value)
from karabogui.singletons.mediator import Mediator
from karabogui.testing import (
    GuiTestCase, access_level, click_button, get_all_props_schema, singletons)

from ..configurationpanel import CONFIGURATION_PAGE, ConfigurationPanel
from ..panel_info import create_configurator_info

CONFIG_PANEL_PATH = "karabogui.panels.configurationpanel.ConfigurationPanel"
SHOW_NOT_LOADED_PROPERTIES_PATH = (CONFIG_PANEL_PATH
                                   + "._show_not_loaded_properties")

DEFAULT_VALUES = {
    "uintProperty": 2,
    "floatProperty": 2.0,
    "stringProperty": "foo",
    "vectorUint": [1, 2, 3],
    "vectorBool": [True, False, True],
    "vectorHash": [Hash('uintProperty', 1,
                        'stringProperty', 'spam',
                        'floatProperty', 5.0)],
    "vectorHashReadOnlyColumn": [Hash('uintProperty', 1,
                                      'floatProperty', 5.0)],
    "vectorHashInitOnlyColumn": [Hash('uintProperty', 1,
                                      'floatProperty', 5.0)]}


class TableRow(Configurable):
    uintProperty = UInt32(defaultValue=1)
    stringProperty = String(defaultValue="spam")
    floatProperty = Float(defaultValue=5.0)


class TableRowEmpty(Configurable):
    uintProperty = UInt32()
    stringProperty = String()
    floatProperty = Float()


class TableRowReadOnly(Configurable):
    uintProperty = UInt32(defaultValue=1, accessMode=AccessMode.READONLY)
    floatProperty = Float(defaultValue=5.0)


class TableRowInitOnly(Configurable):
    uintProperty = UInt32(defaultValue=1, accessMode=AccessMode.INITONLY)
    floatProperty = Float(defaultValue=5.0)


class Object(Configurable):
    state = String(defaultValue=State.ON)

    uintProperty = UInt32()
    floatProperty = Float()
    stringProperty = String()

    vectorUint = VectorUInt32()
    vectorBool = VectorBool()

    uintReadOnly = UInt32(accessMode=AccessMode.READONLY)
    uintInitOnly = UInt32(accessMode=AccessMode.INITONLY)

    internal = Bool(
        defaultValue=True,
        assignment=Assignment.INTERNAL)

    # -----------------------------------------------------------------------
    # Disable the MDL sanitization for tables in tests

    def new_sanitize(schema, ro):
        return schema

    with patch('karabo.native.schema.descriptors.sanitize_table_schema',
               new=new_sanitize):
        vectorHash = VectorHash(rows=TableRow)
        vectorHashReadOnly = VectorHash(rows=TableRow,
                                        accessMode=AccessMode.READONLY)
        vectorHashInitOnly = VectorHash(rows=TableRow,
                                        accessMode=AccessMode.INITONLY)
        vectorHashReadOnlyColumn = VectorHash(rows=TableRowReadOnly)
        vectorHashInitOnlyColumn = VectorHash(rows=TableRowInitOnly)
        vectorHashEmpty = VectorHash(TableRowEmpty)


class TestSetProxyConfiguration(GuiTestCase):

    def setUp(self):
        super().setUp()
        # Setup online device proxy
        online_binding = build_binding(Object().getDeviceSchema())
        apply_configuration(Hash(DEFAULT_VALUES), online_binding)
        self.online_device_proxy = DeviceProxy(binding=online_binding,
                                               server_id='Test',
                                               status=ProxyStatus.ONLINE)
        # Setup offline device proxy
        offline_binding = build_binding(Object().getDeviceSchema())
        apply_configuration(Hash(DEFAULT_VALUES), offline_binding)
        self.offline_device_proxy = ProjectDeviceProxy(
            binding=offline_binding, server_id='Test',
            status=ProxyStatus.OFFLINE)

        self.config_panel = ConfigurationPanel()

    def tearDown(self):
        super().tearDown()
        self.config_panel._reset_panel()
        self.config_panel.destroy()
        self.config_panel = None

    def test_uint32_binding(self):
        self._assert_default_value("uintProperty")

        # -- Check with valid values -- #
        self._assert_config(uintProperty={"value": 22, "valid": True})
        # Check with a cast-able string
        self._assert_config(
            uintProperty={"value": "22",
                          "valid": True,
                          "casted": True})

        # -- Check with invalid values -- #
        self._assert_config(uintProperty={"value": -2, "valid": False})
        # Check with invalid uncast-able string
        self._assert_config(uintProperty={"value": "foo", "valid": False})
        # Check with floats. Should not pass, even with an integer-like value
        # XXX: Should it really not pass?
        self._assert_config(uintProperty={"value": 2.0, "valid": False})

    def test_float_binding(self):
        self._assert_default_value("floatProperty")

        # -- Check with valid values -- #
        self._assert_config(floatProperty={"value": 3.0, "valid": True})
        self._assert_config(floatProperty={"value": True, "valid": True})
        # Check with a cast-able string
        self._assert_config(
            floatProperty={"value": "3.33", "valid": True, "casted": True})

        # -- Check with invalid values -- #
        self._assert_config(floatProperty={"value": "foo", "valid": False})
        self._assert_config(floatProperty={"value": [1, 2, 3], "valid": False})

    def test_string_binding(self):
        self._assert_default_value("stringProperty")

        # -- Check with valid values -- #
        self._assert_config(
            stringProperty={"value": "bar", "valid": True})
        # Check with invalid but cast-able value
        self._assert_config(
            stringProperty={"value": 1, "valid": True, "casted": True})

        # -- Check with invalid values -- #
        self._assert_config(
            stringProperty={"value": None, "valid": False})

    def test_vector_uint32_binding(self):
        self._assert_default_value("vectorUint")

        # -- Check with valid values -- #
        self._assert_config(vectorUint={"value": [4, 5, 6], "valid": True})
        # Check with vector floats with integer values
        self._assert_config(vectorUint={"value": [4.0, 5.0, 6.0],
                                        "valid": True})

        # -- Check with invalid values -- #
        # Check with list that contains negative integers
        self._assert_config(vectorUint={"value": [-1, -2, 3], "valid": False})
        # Check with list that contains floats
        self._assert_config(vectorUint={"value": [4.4, 5.5, 6.6],
                                        "valid": False})

    def test_vector_bool_binding(self):
        self._assert_default_value("vectorBool")

        # -- Check with valid values -- #
        self._assert_config(
            vectorBool={"value": [False], "valid": True})
        # Check with vectors of 0s and 1s
        self._assert_config(
            vectorBool={"value": [0, 1, 1], "valid": True})

        # -- Check with invalid values -- #
        self._assert_config(vectorBool={"value": [1, 2, 3], "valid": False})
        self._assert_config(vectorBool={"value": "foo", "valid": False})

    def test_valid_vector_hash(self):
        # Check setting the default value again.. It shouldn't trigger change.
        self._assert_default_value("vectorHash")

        # Check with only one value changed. Should trigger a change.
        value = [Hash('uintProperty', 2,  # this changed
                      'stringProperty', 'spam',
                      'floatProperty', 5.0)]
        self._assert_config(vectorHash={"value": value, "valid": True})

        # Check with values as strings. Should still be a valid change.
        value = [Hash('uintProperty', '2',
                      'stringProperty', 'spam',
                      'floatProperty', '5.0')]
        self._assert_config(
            vectorHash={"value": value, "valid": True, "casted": True})

        # Reordering the default value is a valid input but should not
        # trigger change
        value = [Hash('stringProperty', 'spam',
                      'uintProperty', 1,
                      'floatProperty', 5.0)]
        self._assert_config(
            vectorHash={"value": value, "valid": True, "changed": False})

        # Check multiple rows: one reordered Hash, one new row
        value = [Hash('stringProperty', 'spam',
                      'uintProperty', 1,
                      'floatProperty', 5.0),
                 Hash('uintProperty', 4,
                      'stringProperty', 'foobarbar',
                      'floatProperty', 5.6)]
        self._assert_config(
            vectorHash={"value": value, "valid": True})

        # Check additional columns that are not on the schema. It is ignored.
        # We use the default value as control, it shouldn't trigger change.
        value = [Hash('stringProperty', 'spam',
                      'uintProperty', 1,
                      'floatProperty', 5.0,
                      'nonExistentProperty', 0)]
        self._assert_config(
            vectorHash={"value": value, "valid": True, "changed": False})

        # Check comparison with empty vector hash
        value = [Hash('uintProperty', 1,  # this changed
                      'stringProperty', 'spam',
                      'floatProperty', 5.0)]
        self._assert_config(vectorHashEmpty={"value": value, "valid": True})

    def test_invalid_vector_hash(self):
        # Check with only one invalid value changed.
        # Should not trigger change as the rest are default values
        value = [Hash('uintProperty', -2,  # invalid value for a uint
                      'stringProperty', 'spam',
                      'floatProperty', 5.0)]
        self._assert_config(
            vectorHash={"value": value, "valid": False, "changed": False})

        # Check with changes with one invalid value
        value = [Hash('uintProperty', -2,  # invalid value for a uint
                      'stringProperty', 'foo',  # changed
                      'floatProperty', 5.5)]  # changed
        self._assert_config(
            vectorHash={"value": value, "valid": False, "changed": False})

        # Check with changes with all invalid value
        value = [Hash('uintProperty', -2,  # invalid value for a uint
                      'stringProperty', None,  # invalid for a string
                      'floatProperty', 'foo')]  # invalid for a float
        self._assert_config(
            vectorHash={"value": value, "valid": False, "changed": False})

        # Check comparison with empty vector hash
        value = [Hash('uintProperty', -2,  # invalid value for a uint
                      'stringProperty', 'spam',
                      'floatProperty', 5.0)]
        self._assert_config(
            vectorHashEmpty={"value": value, "valid": False, "changed": False})

    def test_readonly_properties(self):
        # Check with uint read only value. Shouldn't trigger change
        self._assert_config(
            uintReadOnly={"value": 22, "valid": True, "changed": False})

        # Check with a readonly table element. Shouldn't trigger changes.
        value = [Hash('uintProperty', 4,  # new values
                      'floatProperty', 5.6)]  # changed, and valid
        self._assert_config(
            vectorHashReadOnly={"value": value,
                                "valid": True,
                                "changed": False})

        # Check with a reconfigurable table element with a read-only column
        # by setting a changed data on the read-only column and unchanged on
        # the reconfigurable column. This WILL trigger changes!
        value = [Hash('uintProperty', 4,  # changed, but read only
                      'floatProperty', 5.0)]  # not changed
        self._assert_config(
            vectorHashReadOnlyColumn={"value": value,
                                      "valid": True,
                                      "changed": True})

        # Check with a reconfigurable table element with a read-only column
        # by setting a changed data on both the read-only and reconfigurable
        # Should trigger changes.
        value = [Hash('uintProperty', 4,  # changed, but read only
                      'floatProperty', 6.0)]  # changed
        self._assert_config(
            vectorHashReadOnlyColumn={"value": value,
                                      "valid": True,
                                      "changed": True})

        # Check comparison with incomplete vector hash (without default value).
        value = [Hash('uintProperty', 1,  # this changed
                      'floatProperty', 5.0)]
        self._assert_config(vectorHashEmpty={"value": value,
                                             "valid": False,
                                             "changed": False})

    def test_initonly_properties(self):
        """In this test, the behavior of offline and online configurations are
           different. We have to separate the validation and provide
           corresponding expectations."""

        # Check uint property with initonly
        self._assert_online_config(
            uintInitOnly={"value": 22, "valid": True, "changed": False})
        self._assert_offline_config(
            uintInitOnly={"value": 22, "valid": True, "changed": True})

        # Check assignment internal with bool
        # -> changed when online and reconfigurable
        # -> not changed when offline and reconfigurable
        self._assert_online_config(
            internal={"value": False, "valid": True, "changed": True})
        self._assert_offline_config(
            internal={"value": False, "valid": True, "changed": False})

        # Check with an init-only table element. Shouldn't trigger changes on
        # online device but should trigger change on offline device.
        value = [Hash('uintProperty', 4,  # new value
                      'stringProperty', 'spam',
                      'floatProperty', 5.0)]  # new value
        self._assert_online_config(
            vectorHashInitOnly={"value": value,
                                "valid": True,
                                "changed": False})
        self._assert_offline_config(
            vectorHashInitOnly={"value": value,
                                "valid": True,
                                "changed": True})

        # Check with a reconfigurable table element with a init-only column
        # by setting a changed data on the init-only column and unchanged on
        # the reconfigurable column. Shouldn't trigger changes on online device
        # but should change on offline device.
        value = [Hash('uintProperty', 4,  # new value
                      'stringProperty', 'spam',
                      'floatProperty', 5.0)]  # new value
        self._assert_online_config(
            vectorHashInitOnly={"value": value,
                                "valid": True,
                                "changed": False})
        self._assert_offline_config(
            vectorHashInitOnly={"value": value,
                                "valid": True,
                                "changed": True})

        # Check with a reconfigurable table element with a init-only column
        # by setting a changed data on both the init-only and reconfigurable
        # column Should trigger changes both on online and offline devices.
        value = [Hash('uintProperty', 4,  # changed, but init only
                      'floatProperty', 6.0)]  # not changed
        self._assert_online_config(
            vectorHashInitOnlyColumn={"value": value,
                                      "valid": True,
                                      "changed": True})
        self._assert_offline_config(
            vectorHashInitOnlyColumn={"value": value,
                                      "valid": True,
                                      "changed": True})

    def test_multiple_properties(self):
        # Check both valid values
        self._assert_config(
            uintProperty={"value": -2, "valid": False},
            floatProperty={"value": 3.0, "valid": True})

        # Check one invalid values
        self._assert_config(
            uintProperty={"value": 22, "valid": True},
            floatProperty={"value": "foo", "valid": False})

        # Check two invalid values
        self._assert_config(
            uintProperty={"value": 22, "valid": True},
            floatProperty={"value": "foo", "valid": False})

    # -----------------------------------------------------------------------
    # Assertions

    def _assert_config(self, **paths):
        self._assert_online_config(**paths)
        self._assert_offline_config(**paths)

    def _assert_online_config(self, **paths):
        # Set up config and changed values
        config, changed, validity = {}, {}, {}
        for path, values in paths.items():
            config[path] = values["value"]
            validity[path] = values["valid"]
            changed[path] = values.get("changed", values["valid"])

        # Apply changes
        with patch(SHOW_NOT_LOADED_PROPERTIES_PATH) as mocked_method:
            self._set_proxy_config(self.online_device_proxy, config)

        # Get invalid config
        invalid_config = {}
        for path, valid in validity.items():
            binding = self._get_online_property_binding(path)
            if binding is None:
                continue
            value = config[path]
            if isinstance(binding, VectorHashBinding):
                valid, invalid = validate_table_value(binding, value)
                config[path] = valid
                invalid_config[path] = invalid
            elif validate_value(binding, value) is None:
                invalid_config[path] = value

        # Check if dialog is called
        if all(validity.values()):
            mocked_method.assert_not_called()
        else:
            mocked_method.assert_called_once_with(
                self.online_device_proxy, invalid_config)

        # Check edit value if changed
        for path, value in config.items():
            if not changed[path]:
                value = None
            casted = paths[path].get("casted", False)
            self._assert_edit_value(path, value, casted=casted)

    def _assert_offline_config(self, **paths):
        # Reset to default value
        binding = self.offline_device_proxy.binding
        apply_configuration(Hash(DEFAULT_VALUES), binding)

        # Set up config and changed values
        config, changed, validity = {}, {}, {}
        for path, values in paths.items():
            config[path] = values["value"]
            validity[path] = values["valid"]
            changed[path] = values.get("changed", values["valid"])

        # Apply changes
        with patch(SHOW_NOT_LOADED_PROPERTIES_PATH) as mocked_method:
            self._set_proxy_config(self.offline_device_proxy, config)

        # Get invalid config
        invalid_config = Hash()
        for path, valid in validity.items():
            binding = self._get_offline_property_binding(path)
            if binding is None:
                continue
            value = config[path]
            if isinstance(binding, VectorHashBinding):
                valid, invalid = validate_table_value(
                    binding, value)
                config[path] = valid
                invalid_config[path] = invalid
            elif validate_value(binding, value) is None:
                invalid_config[path] = value

        # Check if dialog is called
        if all(validity.values()):
            mocked_method.assert_not_called()
        else:
            mocked_method.assert_called_once_with(
                self.offline_device_proxy, invalid_config)

        # Check offline value if changed
        for path, value in config.items():
            if not changed[path]:
                value = DEFAULT_VALUES.get(path, Undefined)
            casted = paths[path].get("casted", False)
            self._assert_offline_value(path, value, casted=casted)

    def _assert_default_value(self, path):
        # Check with unchanged values
        self._assert_config(**{
            path: {"value": DEFAULT_VALUES[path],
                   "valid": True, "changed": False}
        })

    def _assert_edit_value(self, path, value, casted=False):
        edit_value = self._get_property_proxy(path).edit_value
        self._assert_value(actual=edit_value, expected=value, casted=casted)

    def _assert_offline_value(self, path, value, casted=False):
        actual_value = self._get_offline_property_binding(path).value
        self._assert_value(actual=actual_value, expected=value, casted=casted)

    def _assert_value(self, actual, expected, casted=False):
        if np.isscalar(expected):
            if casted and actual is not None:
                expected = type(actual)(expected)
            assert actual == expected
        elif isinstance(actual, HashList):
            for actual_row, expected_row in zip(actual, expected):
                for path, actual_value in actual_row.items():
                    expected_value = expected_row[path]
                    self._assert_value(actual_value, expected_value, casted)
        else:
            np.testing.assert_array_equal(actual, expected)

    # -----------------------------------------------------------------------
    # Helper functions

    def _set_proxy_config(self, proxy, config):
        self.config_panel._show_configuration(proxy)
        self.config_panel._set_proxy_configuration(proxy, Hash(config))

    def _get_property_proxy(self, path):
        stacked_widget = self.config_panel._stacked_tree_widgets
        editor = stacked_widget.widget(CONFIGURATION_PAGE)
        model = editor.model()
        return model.property_proxy(path)

    def _get_online_property_binding(self, path):
        return self.online_device_proxy.get_property_binding(path)

    def _get_offline_property_binding(self, path):
        return self.offline_device_proxy.get_property_binding(path)


def test_configuration_panel_info(gui_app):
    schema = get_all_props_schema()
    binding = build_binding(schema)

    assert binding.value.a.value is Undefined
    assert binding.value.b.value is Undefined

    apply_default_configuration(binding)
    assert binding.value.a.value
    assert binding.value.b.value == "c"
    assert binding.value.f.value is Undefined

    mediator = Mediator()
    with singletons(mediator=mediator):
        proxy = DeviceProxy(binding=binding,
                            server_id="Test",
                            status=ProxyStatus.ONLINE)

        panel = ConfigurationPanel()
        # XXX: Make to see all
        panel._switch_mode_toggled(True)
        panel.model().root = proxy
        panel._set_configuration(proxy)
        # Make sure the extracted default conversion is minimal
        # It should include properties with default values,
        # options, or node types
        with access_level(AccessLevel.EXPERT):
            config = create_configurator_info(panel)

        assert isinstance(config, Hash)

        default_props = ("a", "b", "h1")
        for prop in default_props:
            assert config[prop] != "<undefined>"

        # List and Choice of Node and not present
        assert "i1" not in config
        assert "j1" not in config

        # Slot k1 is not present
        assert "k1" not in config

        # m has no default value, but present with Undefined
        assert "m" in config
        assert config["m"] == "<undefined>"

        assert "f" in config
        assert config["f"] == "<undefined>"


def test_configuration_from_past(gui_app, mocker):

    panel = ConfigurationPanel()
    assert not panel.ui_history_config.isVisible()

    proxy_binding = build_binding(Object().getDeviceSchema())
    apply_configuration(Hash(DEFAULT_VALUES), proxy_binding)
    proxy = DeviceProxy(
        binding=proxy_binding, server_id='Test', status=ProxyStatus.ONLINE)
    panel._set_configuration(proxy)
    assert panel.ui_history_config.isVisible()

    # Make sure the button shows the ConfigurationFromPastDialog
    toolbar = panel.ui_history_config.parent()
    button = toolbar.widgetForAction(panel.ui_history_config)
    assert button.toolTip() == "Runtime configuration"
    mock_dialog = mocker.patch(
        "karabogui.panels.configurationpanel.ConfigurationFromPastDialog")
    click_button(button)
    assert mock_dialog().show.call_count == 1

    # Test the visibility of toolbutton.
    # Offline Device
    proxy = DeviceProxy(
        binding=proxy_binding, status=ProxyStatus.OFFLINE)
    panel._set_configuration(proxy)
    assert panel.ui_history_config.isVisible()

    # Device class proxy
    proxy = DeviceClassProxy(binding=proxy_binding,)
    panel._set_configuration(proxy)
    assert not panel.ui_history_config.isVisible()

    # Project Device Proxy
    proxy = ProjectDeviceProxy(binding=proxy_binding)
    panel._set_configuration(proxy)
    assert panel.ui_history_config.isVisible()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np

from karabo.common.alarm_conditions import AlarmCondition
from karabo.common.states import State

# Toolbar
# --------------------------------------------------------------------------

BUTTON_SIZE = 32
ICON_SIZE = 20

# Global Constants
# --------------------------------------------------------------------------

AXIS_ITEMS = ["top", "bottom", "left", "right"]
AXIS_X = ["top", "bottom"]
AXIS_Y = ["left", "right"]

WIDGET_MIN_HEIGHT = 200  # pixels
WIDGET_MIN_WIDTH = 300

WIDGET_HEIGHT_HINT = 250  # pixels
WIDGET_WIDTH_HINT = 450

# Plot Constants
# --------------------------------------------------------------------------

DEFAULT_BAR_WIDTH = 0.1
DEFAULT_PEN_WIDTH = 1.0

SYMBOL_SIZE = 5
DEFAULT_SYMBOL = 'o'
EMPTY_SYMBOL_OPTIONS = {'symbol': None, 'symbolSize': None,
                        'symbolPen': None, 'symbolBrush': None}
EMPTY_BAR = np.zeros(shape=(10,), dtype=np.int64)

CHECK_ACTIONS = ['x_grid', 'y_grid', 'x_log', 'y_log',
                 'x_invert', 'y_invert']
RANGE_ACTIONS = ['axes', 'x_range', 'y_range']
VIEW_ACTIONS = ['view']
ACTION_ITEMS = CHECK_ACTIONS + RANGE_ACTIONS + VIEW_ACTIONS

# Items required for the state trendline!
ALL_STATES = [
    State.NORMAL,
    State.UNKNOWN,
    State.INIT,
    State.ERROR,
    State.INTERLOCKED,
    State.DISABLED,
    State.CHANGING,
    State.MOVING,
    State.OFF,
    State.ON,
    State.STOPPED,
    State.RUNNING,
    State.ACQUIRING,
    State.PROCESSING,
    State.HOMING,
    State.ACTIVE,
    State.PASSIVE,
    State.OPENED,
    State.CLOSED,
    State.PAUSED,
    State.OPENING,
    State.CLOSING,
    State.STARTED,
    State.INSERTED,
    State.MONITORING,
    State.IGNORING,
    State.INSERTING,
    State.STARTING,
    State.STOPPING,
    State.SEARCHING,
    State.STATIC,
    State.RAMPING_DOWN,
    State.RAMPING_UP,
    State.EXTRACTING,
    State.EXTRACTED,
    State.COOLING,
    State.COOLED,
    State.COLD,
    State.HEATING,
    State.HEATED,
    State.WARM,
    State.INTERLOCK_OK,
    State.INTERLOCK_BROKEN,
    State.PRESSURIZED,
    State.EVACUATED,
    State.EMPTYING,
    State.FILLING,
    State.DISENGAGING,
    State.DISENGAGED,
    State.ENGAGING,
    State.ENGAGED,
    State.SWITCHING_OFF,
    State.SWITCHING_ON,
    State.SWITCHING,
    State.ROTATING,
    State.ROTATING_CNTCLK,
    State.ROTATING_CLK,
    State.MOVING_LEFT,
    State.MOVING_DOWN,
    State.MOVING_BACK,
    State.MOVING_RIGHT,
    State.MOVING_UP,
    State.MOVING_FORWARD,
    State.UNLOCKED,
    State.LOCKED,
    State.INCREASING,
    State.DECREASING,
    State.KNOWN,
]

STATE_INTEGER_MAP = {state.value: i for i, state in enumerate(ALL_STATES)}
INTEGER_STATE_MAP = {i: state.value for i, state in enumerate(ALL_STATES)}

ALARM_INTEGER_MAP = {alarm.value: i for i, alarm in
                     enumerate(AlarmCondition.__members__.values())}
INTEGER_ALARM_MAP = {i: alarm.value for i, alarm in
                     enumerate(AlarmCondition.__members__.values())}


def get_alarm_string(value):
    """Get an alarm condition from a mapped `index`"""
    return INTEGER_ALARM_MAP.get(round(value), "")


def get_state_string(value):
    """Get an state value from a mapped `index`"""
    return INTEGER_STATE_MAP.get(round(value), "")


# Image Constants
# --------------------------------------------------------------------------

TF_SCALING = "scaling"
TF_TRANSLATION = "translation"
TF_FLIPPED = "flipped"
TF_ROTATION = "rotation"

ROTATION_FACTOR = {
    0: (1, 1),
    90: (1, -1),
    180: (-1, -1),
    270: (-1, 1)
}

DEFAULT_SCALE_X = 1.0
DEFAULT_SCALE_Y = 1.0
DEFAULT_OFFSET_X = 0.0
DEFAULT_OFFSET_Y = 0.0

X_AXIS_HEIGHT = 46
Y_AXIS_WIDTH = 75

DEFAULT_UNITS_X = "pixels"
DEFAULT_LABEL_X = "X-axis"
DEFAULT_UNITS_Y = "pixels"
DEFAULT_LABEL_Y = "Y-axis"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
import pytest

from karabo.native import Configurable, Double, Hash, NDArray, VectorFloat
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..vector_hist_graph import VectorHistogramGraph


class VectorObject(Configurable):
    prop = VectorFloat(defaultValue=[2.3, 4.5, 7.9])


class NDArrayObject(Configurable):
    prop = NDArray(
        defaultValue=[2.3, 4.5, 7.9, 1.0, 2.0, 3.0, 4.0, 3.0, 9.0, 10.1],
        shape=(10,),
        dtype=Double)


@pytest.fixture
def vector_hist_setup(gui_app):
    schema = VectorObject.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = VectorHistogramGraph(proxy=proxy)
    controller.create(None)
    assert controller.widget is not None
    yield controller, proxy
    controller.destroy()
    assert controller.widget is None


def test_vector_hist_basics(vector_hist_setup):
    controller, proxy = vector_hist_setup
    controller.model.auto = False
    controller.model.start = 0
    controller.model.stop = 10
    set_proxy_value(proxy, "prop", [1, 2, 2, 3, 3, 3, 3, 8])
    curve = controller._plot
    x, y = curve.getData()
    np.testing.assert_array_almost_equal(
        x, [-0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5])
    np.testing.assert_array_almost_equal(y, [0, 1, 2, 4, 0, 0, 0, 0, 1, 0])


def test_vector_hist_empty(vector_hist_setup):
    controller, proxy = vector_hist_setup
    controller.model.auto = False
    controller.model.start = 0
    controller.model.stop = 10
    set_proxy_value(proxy, "prop", [])
    curve = controller._plot
    x, y = curve.getData()
    assert x is None
    assert y is None


def test_array_hist_basics(gui_app):
    # setup
    schema = NDArrayObject.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = VectorHistogramGraph(proxy=proxy)
    controller.create(None)
    assert controller.widget is not None

    # test body
    controller.model.auto = False
    controller.model.start = 0
    controller.model.stop = 10
    value = np.array([1.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 8.1, 9.2, 10.1],
                     dtype=np.float64)
    array_hash = Hash("type", 20,
                      "data", value.tobytes())
    set_proxy_value(proxy, "prop", array_hash)

    curve = controller._plot
    x, y = curve.getData()
    np.testing.assert_array_almost_equal(
        x, [-0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5])
    np.testing.assert_array_almost_equal(y, [9, 1, 9, 0, 0, 0, 0, 0, 0, 0])

    # teardown
    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path

import numpy as np
from qtpy import uic
from qtpy.QtCore import Qt, Signal, Slot
from qtpy.QtWidgets import QDialog

from karabogui.graph.common.utils import float_to_string
from karabogui.util import SignalBlocker
from karabogui.widgets.range_slider import RangeSlider


def get_decimals(array, offset=2, default=1):
    """Calculates the number of decimals to display with the input number
    e.g., 0.01 -> 0.0100
          10.0 -> 10.0
    """
    array = np.abs(array)
    if not np.all((array > 0) & (array < 1)):
        return default

    decimals = max(-1 * np.floor(np.log10(array)))
    return int(decimals + offset)


class LevelsDialog(QDialog):
    levelsPreview = Signal(object)

    def __init__(self, levels, image_range, auto_levels, limits=None,
                 parent=None):
        super().__init__(parent)
        ui_path = Path(__file__).parent.joinpath("levels_dialog.ui")
        uic.loadUi(ui_path, self)
        self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)
        self.values_widget.setEnabled(not auto_levels)

        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        # Check if autolevel: image levels and range are almost equal.
        # This is with a tolerance of 1%.
        self.automatic_checkbox.setChecked(auto_levels)
        self.automatic_checkbox.stateChanged.connect(self.set_automatic_levels)

        self._image_range = image_range
        self.slider = RangeSlider(Qt.Horizontal, parent=self)
        self.slider_layout.addWidget(self.slider)

        # Set slider and editor widget default values
        min_level, max_level = levels
        min_range, max_range = image_range

        slider_min = min(min_level, min_range)
        slider_max = max(max_level, max_range)
        self.slider.initialize(slider_min, slider_max)

        # Handle floats
        is_float = (isinstance(slider_min, np.floating)
                    or isinstance(slider_max, np.floating))
        decimals = get_decimals([slider_min, slider_max]) if is_float else 0

        # Set decimals
        self.min_spinbox.setDecimals(decimals)
        self.max_spinbox.setDecimals(decimals)

        # Set single steps
        step = 10 ** -(decimals - 1) if decimals else 1
        self.min_spinbox.setSingleStep(step)
        self.max_spinbox.setSingleStep(step)

        if limits is not None:
            min_value, max_value = limits
            self.min_spinbox.setMinimum(min_value)
            self.max_spinbox.setMinimum(min_value)
            self.min_spinbox.setMaximum(max_value)
            self.max_spinbox.setMaximum(max_value)

        default = image_range if auto_levels else levels
        self._set_editor_default(default)

        # Connect signals
        self.slider.sliderMoved.connect(self.sliderMoved)
        self.min_spinbox.valueChanged.connect(self.levelChanged)
        self.max_spinbox.valueChanged.connect(self.levelChanged)

        # Finally select the spinbox
        if not auto_levels:
            self.min_spinbox.setFocus()
            self.min_spinbox.selectAll()

    def _set_editor_default(self, levels):
        """Set the default values of the editor widgets and the slider"""
        min_level, max_level = levels
        decimals = get_decimals(levels)

        self.min_label.setText(float_to_string(min_level, decimals))
        self.max_label.setText(float_to_string(max_level, decimals))
        self.min_spinbox.setValue(min_level)
        self.max_spinbox.setValue(max_level)
        self.slider.setValue(min_level, max_level)

    @Slot()
    def levelChanged(self):
        values = [self.min_spinbox.value(), self.max_spinbox.value()]
        min_level = min(values)
        max_level = max(values)
        with SignalBlocker(self.slider):
            self.slider.setValue(min_level, max_level)
        levels = [min_level, max_level]
        self.levelsPreview.emit(levels)

    @Slot(int, int)
    def sliderMoved(self, low, high):
        self.min_spinbox.setValue(low)
        self.max_spinbox.setValue(high)

    @Slot(int)
    def set_automatic_levels(self, state):
        auto_levels = state == Qt.Checked
        self.values_widget.setEnabled(not auto_levels)
        low, high = self._image_range
        with SignalBlocker(self.min_spinbox, self.max_spinbox, self.slider):
            self.min_spinbox.setValue(low)
            self.max_spinbox.setValue(high)
            self.slider.setValue(low, high)
            if not auto_levels:
                self.min_spinbox.setFocus()
                self.min_spinbox.selectAll()

        levels = None if auto_levels else [low, high]
        self.levelsPreview.emit(levels)

    @property
    def levels(self):
        if self.automatic_checkbox.isChecked():
            return None

        min_level = min([self.min_spinbox.value(), self.max_spinbox.value()])
        max_level = max([self.min_spinbox.value(), self.max_spinbox.value()])
        levels = [min_level, max_level]
        return levels



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from functools import partial

from pyqtgraph import GraphicsLayoutWidget
from qtpy.QtCore import QEvent, QSize, Qt, Signal, Slot
from qtpy.QtGui import QPalette
from qtpy.QtWidgets import QAction, QGridLayout, QWidget

from karabogui import icons
from karabogui.graph.common.api import (
    COLORMAPS, AuxPlots, AxesLabelsDialog, ExportTool, ImageROIController,
    MouseTool, PointCanvas, RectCanvas, ROITool, ToolbarController)
from karabogui.graph.common.const import TF_SCALING, X_AXIS_HEIGHT

from .aux_plots.controller import AuxPlotsController
from .colorbar import ColorBarWidget
from .dialogs.transforms import ImageTransformsDialog
from .legends.scale import ScaleLegend
from .plot import KaraboImagePlot
from .tools.picker import PickerController
from .tools.toolbar import AuxPlotsToolset
from .utils import (
    create_colormap_menu, create_icon_from_colormap, levels_almost_equal)

ROI_CANVAS_MAP = {
    ROITool.DrawRect: (RectCanvas, ROITool.Rect),
    ROITool.DrawCrosshair: (PointCanvas, ROITool.Crosshair)}


class KaraboImageView(QWidget):
    stateChanged = Signal(object)
    toolTipChanged = Signal()

    def __init__(self, parent=None):
        """ The main image view widget for a single ImagePlot

        Acts as a container of the graphics layout, the image plotitem,
        the toolbar and other tooling features.
        """
        super().__init__(parent)

        # Main layout to organize
        layout = QGridLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.image_layout = GraphicsLayoutWidget()
        self.layout().addWidget(self.image_layout, 0, 0, 1, 1)

        # Image can have transparent background!
        palette = self.image_layout.palette()
        palette.setColor(QPalette.Background, Qt.transparent)
        self.image_layout.setPalette(palette)

        # Add our basic plotItem to this widget
        self.plotItem = KaraboImagePlot(parent=self.image_layout.ci)
        self.plotItem.imageLevelsChanged.connect(self._image_levels_changed)
        self.plotItem.imageItem.sigImageChanged.connect(self._image_changed)

        # row, col, row_span, col_span
        self.image_layout.addItem(self.plotItem, 1, 1, 1, 1)
        self.image_layout.ci.layout.setRowStretchFactor(1, 3)
        self.image_layout.ci.layout.setColumnStretchFactor(1, 3)
        # Note: Leave a little bottom margin for the color bar
        self.image_layout.ci.layout.setContentsMargins(0, 0, 0, 5)

        # Our tooling instances
        self._aux_plots = None
        self._colorbar = None
        self._canvas = None
        self._picker = None

        self.toolbar = None
        self.roi = None

        self._colormap_action = None
        self._apply_action = None
        self._scale_legend = None
        self._save_levels_action = None

        self.configuration = {}

    # -----------------------------------------------------------------------
    # Public tool methods

    def add_toolbar(self):
        """Enable the standard toolbar on the widget

        This is the finalizer of the `KaraboImageView` widget. All signals
        should be connected here!
        """
        if self.toolbar is not None:
            raise ValueError('Toolbar already added')

        # Instantiate toolbar, has a default mouse mode toolset
        self.toolbar = tb = ToolbarController(parent=self)
        tb.toolsets[MouseTool].on_trait_change(
            self.set_mouse_mode, "current_tool")

        # Add the optional mouse mode: picker tool
        if self._picker is not None:
            tb.add_tool(MouseTool.Picker)

        # Add Aux Plots toolset
        if self._aux_plots is not None:
            tb.register_toolset(toolset=AuxPlots, klass=AuxPlotsToolset)
            aux_plots_toolset = tb.add_toolset(AuxPlots)
            aux_plots_toolset.on_trait_change(self.show_aux_plots,
                                              "current_tool")

        # Add ROI toolset
        if self.roi is not None:
            roi_toolset = tb.add_toolset(ROITool)
            roi_toolset.on_trait_change(self._activate_roi_tool,
                                        "current_tool")
            # Connect signals from the items
            self.roi.selected.connect(roi_toolset.check)
            self.roi.removed.connect(roi_toolset.uncheck_all)

        # Add Export toolset
        export_toolset = tb.add_toolset(ExportTool)
        export_toolset.on_trait_change(self.plotItem.export, "current_tool")

        # row, col, row_span, col_span
        self.layout().addWidget(tb.widget, 0, 1, 1, 1)

        return self.toolbar

    def add_colorbar(self):
        """Enable the standard colorbar of this widget"""
        if self._colorbar is None:
            plotItem = self.plotItem

            image = plotItem.imageItem
            self._colorbar = ColorBarWidget(image, parent=plotItem)

            top_axis_checked = plotItem.getAxis("top").style["showValues"]
            top_margin = X_AXIS_HEIGHT * top_axis_checked

            bot_axis_checked = plotItem.getAxis("bottom").style["showValues"]
            bottom_margin = X_AXIS_HEIGHT * bot_axis_checked

            self._colorbar.set_margins(top=top_margin, bottom=bottom_margin)
            self._colorbar.levelsChanged.connect(
                plotItem.set_image_levels)

            self.image_layout.addItem(self._colorbar, row=1, col=2)
            self.image_layout.ci.layout.setColumnStretchFactor(2, 1)

            self.add_colormap_action()

            if self._save_levels_action is None:
                self._save_levels_action = QAction(self)
                self._save_levels_action.setIcon(icons.apply)
                self._save_levels_action.setIconText("Save Color Levels")
                self._save_levels_action.triggered.connect(
                    self._save_color_levels)
                self.addAction(self._save_levels_action)

        return self._colorbar

    def remove_colorbar(self):
        if self._colorbar is not None:
            self.image_layout.removeItem(self._colorbar)
            self.image_layout.ci.layout.setColumnStretchFactor(1, 3)
            self._colorbar.deleteLater()
            self._colorbar = None

            self.remove_colormap_action()

    def add_roi(self, enable=True):
        if enable and self.roi is None:
            # Initialize ROI controller
            self.roi = ImageROIController(self.plotItem)
            if self._apply_action is None:
                self._apply_action = QAction(self)
                self._apply_action.setIcon(icons.apply)
                self._apply_action.setIconText('Set ROI and Aux')
                self._apply_action.triggered.connect(
                    self._set_roi_configuration)
                self.addAction(self._apply_action)

            # Set the minimum size of the plot without rescaling problems
            # from the ROI items (e.g., zooming out of infinity when it is
            # collapsed), which is 300 x 300.
            self.plotItem.setMinimumWidth(300)
            self.plotItem.setMinimumHeight(300)

        elif not enable and self.roi is not None:
            # Remove ROI items and destroy controller
            self.roi.remove_all()
            self.roi.destroy()
            self.roi = None
            self._apply_action.disconnect()
            self.removeAction(self._apply_action)
            self._apply_action = None

        return self.roi

    def add_picker(self, enable=True):
        if enable and self._picker is None:
            self._picker = PickerController(self.plotItem, parent=self)
        elif not enable and self._picker is not None:
            self._picker.destroy()
            self._picker = None

        return self._picker

    def add_aux(self, plot=None, **config):
        """Add a auxiliary plots to the ImageView"""
        if plot is not None:
            if self._aux_plots is None:
                # Create an instance of the aux plots controller with the klass
                self._aux_plots = AuxPlotsController(
                    image_layout=self.image_layout)  # noqa

                self.plotItem.imageAxesChanged.connect(self._set_axes_to_aux)

                # If ROI exists
                if self.roi is not None:
                    self.roi.updated.connect(self._aux_plots.process)

            controller = self._aux_plots.add(plot, **config)
            controller.link_view(self.plotItem.vb)

        else:
            if plot is AuxPlots.NoPlot:
                self._aux_plots.current_plot.link_view(None)
                self._aux_plots.clear()

                # If ROI exists
                if self.roi is None:
                    self.roi.updated.disconnect(self._aux_plots.analyze)

                self.plotItem.imageAxesChanged.disconnect(
                    self._set_axes_to_aux)

        return self._aux_plots

    def enable_aux(self):
        if self.toolbar is None:
            return

        # Enable toolbar buttons
        toolset = self.toolbar.toolsets.get(AuxPlots)
        if toolset is not None:
            for button in toolset.buttons.values():
                button.setEnabled(True)

    def disable_aux(self):
        if self._aux_plots is None:
            return

        # Hide existing aux plot if showing
        self.restore({"aux_plots": AuxPlots.NoPlot})

        # Disable toolbar buttons
        if self.toolbar is not None:
            toolset = self.toolbar.toolsets.get(AuxPlots)
            if toolset is not None:
                for aux_plot, button in toolset.buttons.items():
                    if aux_plot == self._aux_plots.current_plot:
                        button.setChecked(False)
                    button.setEnabled(False)

    def add_colormap_action(self, cmap="none"):
        if self._colormap_action is None:
            menu = create_colormap_menu(COLORMAPS, cmap, self.set_colormap)
            self._colormap_action = QAction(self)
            self._colormap_action.setIconText("Colormap")
            self._colormap_action.setMenu(menu)
            self.addAction(self._colormap_action)

        return self._colormap_action

    def remove_colormap_action(self):
        if self._colormap_action is not None:
            self._colormap_action.deleteLater()
            self.removeAction(self._colormap_action)
            self._colormap_action = None

    def add_axes_labels_dialog(self):
        axes_action = QAction("Axes Labels", self)
        axes_action.triggered.connect(self._show_labels_dialog)
        self.addAction(axes_action)

        return axes_action

    def add_transforms_dialog(self):
        transforms_action = QAction("Transformations", self)
        transforms_action.triggered.connect(self._show_transforms_dialog)
        self.addAction(transforms_action)

        return transforms_action

    # -----------------------------------------------------------------------
    # Public Interface

    def plot(self):
        return self.plotItem

    def set_colormap(self, colormap, update=True):
        self.plotItem.set_colormap(colormap)
        if self._colorbar is not None:
            self._colorbar.set_colormap(colormap)
        if self._aux_plots is not None:
            self._aux_plots.set_config(AuxPlots.Histogram, colormap=colormap)

        # NOTE: We might have a colormap action without colorbar!
        if self._colormap_action is not None:
            self._colormap_action.setIcon(create_icon_from_colormap(colormap))

        if self._picker is not None:
            self._picker.update()

        if update:
            config = {'colormap': colormap}
            self.configuration.update(**config)
            self.stateChanged.emit(config)

    def add_widget(self, widget, row, col, row_span=1, col_span=1):
        self.layout().addWidget(widget, row, col, row_span, col_span)

    def add_layout(self, layout, row, col, row_span=1, col_span=1):
        self.layout().addLayout(layout, row, col, row_span, col_span)

    def restore(self, configuration):
        """This method is responsible for restoring the state of all widgets.

        Ideally this should be called at the end of the widget setup.
        """
        # Update internal configuration
        self.configuration.update(**configuration)

        # Restore colormap
        colormap = configuration.get('colormap')
        if colormap is not None:
            self.set_colormap(colormap, update=False)
            # We check the action as well.
            if self._colormap_action is not None:
                for cmap_action in self._colormap_action.menu().actions():
                    if cmap_action.text() == colormap:
                        cmap_action.setChecked(True)

        # Restore labels
        x_label = configuration.get('x_label')
        x_units = configuration.get('x_units')
        if x_label is not None or x_units is not None:
            self.set_label(axis=0,
                           text=x_label or '',
                           units=x_units or '')
        y_label = configuration.get('y_label')
        y_units = configuration.get('y_units')
        if y_label is not None or y_units is not None:
            self.set_label(axis=1,
                           text=y_label or '',
                           units=y_units or '')

        # Restore transforms
        transforms = {k: v for k, v in configuration.items()
                      if k in ['x_scale', 'y_scale',
                               'x_translate', 'y_translate',
                               'aspect_ratio']}
        if transforms:
            kwargs = transforms
            kwargs.update({'default': True})
            self.plotItem.set_transform(**kwargs)
            # Restore scale legend
            show_legend = configuration.get('show_scale')
            self._show_scale_legend(show_legend)
            self._update_scale_legend()

        # Restore ROIs
        current_roi_tool = configuration.get('roi_tool')
        roi_items = configuration.get('roi_items')
        if self.roi is not None:
            # Restore ROI items
            if roi_items is not None:
                self.roi.remove_all()
                for roi_data in roi_items:
                    roi_type = roi_data['roi_type']
                    name = roi_data['name']
                    if roi_type == ROITool.Rect:
                        x, y, w, h = (roi_data['x'], roi_data['y'],
                                      roi_data['w'], roi_data['h'])
                        pos = x, y
                        size = w, h
                    elif roi_type == ROITool.Crosshair:
                        pos = (roi_data['x'], roi_data['y'])
                        size = None
                    else:
                        continue

                    self.roi.add(roi_type, pos, size=size, name=name)

            # Show ROI tool
            if current_roi_tool is not None:
                self.roi.show(current_roi_tool)

        # Restore auxiliary plots
        aux_plots_class = configuration.get('aux_plots')
        if self._aux_plots is not None and aux_plots_class is not None:
            self.show_aux_plots(aux_plots_class)

        # Restore toolbar state
        if self.toolbar is not None:
            # Restore ROI state
            toolset = self.toolbar.toolsets.get(ROITool)
            if toolset is not None and current_roi_tool is not None:
                toolset.check(current_roi_tool)

            # Restore aux plots state
            toolset = self.toolbar.toolsets.get(AuxPlots)
            if toolset is not None and aux_plots_class is not None:
                toolset.check(aux_plots_class)

        # Restore color levels if available (None, [])
        color_levels = configuration.get('color_levels')
        if not color_levels:
            color_levels = None
        self.plotItem.set_image_levels(color_levels)

    def deactivate_roi(self):
        """Deactivate the region of interest"""
        if self.roi is not None:
            for name, roi_type in self.roi.roi_items.items():
                for roi in roi_type:
                    roi.translatable = False

            if self.toolbar is not None:
                for roi_type in [ROITool.Rect, ROITool.Crosshair]:
                    buttons = self.toolbar.buttons[roi_type.name]
                    menu = buttons.menu()
                    action = menu.actions()[1]
                    menu.removeAction(action)

    # -----------------------------------------------------------------------
    # Qt Slots

    @Slot()
    def _image_changed(self):
        """Unified handle of image changes. This is to avoid race conditions.
           Will move the picker and ROI slots here as well"""

        # Check effective colorbar levels. In the ideal world, the effective
        # levels should not be bounded to the colorbar.
        if self._colorbar is not None:
            image_levels = self.plotItem.imageItem.levels
            if not levels_almost_equal(self._colorbar.levels, image_levels):
                self._colorbar.set_levels(image_levels)
                if self._aux_plots is not None:
                    self._aux_plots.set_config(plot=AuxPlots.Histogram,
                                               levels=image_levels)

        if self.roi is not None:
            self.roi.update()

    @Slot(object)
    def _image_levels_changed(self, levels):
        """Unified handle of image level changes."""

        if self._picker is not None:
            self._picker.update()

        if self._aux_plots is not None and self._colorbar is not None:
            levels = levels if levels is not None else self._colorbar.levels
            self._aux_plots.set_config(plot=AuxPlots.Histogram, levels=levels)

    @Slot(MouseTool)
    def set_mouse_mode(self, mode):
        self.plotItem.vb.set_mouse_tool(mode)
        if self._picker is not None:
            self._picker.activate(mode is MouseTool.Picker)

    @Slot(AuxPlots)
    def show_aux_plots(self, plot_type):
        """Hides/shows the auxiliar plots set for this controller"""
        if self._aux_plots is not None:
            self._aux_plots.current_plot = plot_type

        if self.roi is not None:
            self.roi.enable_updates(plot_type != AuxPlots.NoPlot)

    @Slot()
    def _show_labels_dialog(self):
        config, result = AxesLabelsDialog.get(self.configuration,
                                              parent=self)

        if not result:
            return

        self.set_label(axis=0,
                       text=config["x_label"],
                       units=config["x_units"])
        self.set_label(axis=1,
                       text=config["y_label"],
                       units=config["y_units"])

        self._update_scale_legend()

        self.configuration.update(**config)
        self.stateChanged.emit(config)

    def set_label(self, axis, text='', units=''):
        """Public interface similar with to KaraboPlotView"""
        self.plotItem.set_label(axis, text, units)

        # Modify colorbar top margin whether labels are present/absent
        if axis == 0 and self._colorbar is not None:
            margin = X_AXIS_HEIGHT
            if text == '' and units == '':
                margin -= 20
            self._colorbar.set_margins(top=margin)

    @Slot()
    def _show_transforms_dialog(self):
        transform = self.plotItem.axes_transform
        aspect_ratio = self.plotItem.aspect_ratio
        show_legend = self._scale_legend is not None

        config, result = ImageTransformsDialog.get(
            transform, aspect_ratio, show_legend, parent=self)

        if not result:
            return

        self.plotItem.set_transform(x_scale=config["x_scale"],
                                    y_scale=config["y_scale"],
                                    x_translate=config["x_translate"],
                                    y_translate=config["y_translate"],
                                    aspect_ratio=config["aspect_ratio"])

        self._show_scale_legend(show=config["show_scale"])

        self._update_scale_legend()

        self.configuration.update(**config)
        self.stateChanged.emit(config)

    @Slot()
    def _set_roi_configuration(self):
        config = {}
        aux_plots_tool = self.toolbar.toolsets[AuxPlots].current_tool
        if aux_plots_tool is not None:
            config['aux_plots'] = aux_plots_tool

        config['roi_tool'] = self.roi.current_tool
        items = []
        # Save ROI information
        for tool, roi_items in self.roi.roi_items.items():
            # Each tool has multiple roi objects
            for roi in roi_items:
                traits = {'roi_type': tool, "name": roi.name}
                if tool == ROITool.Crosshair:
                    x, y = roi.coords
                    traits.update({'x': x, 'y': y})
                else:
                    x, y, w, h = roi.coords
                    traits.update({'x': x, 'y': y, 'w': w, 'h': h})
                items.append(traits)

        config['roi_items'] = items
        self.configuration.update(**config)
        self.stateChanged.emit(config)

    @Slot()
    def _set_axes_to_aux(self):
        self._aux_plots.set_axes(*self.plotItem.transformed_axes)

    @Slot()
    def _save_color_levels(self):
        image_item = self.plotItem.imageItem
        levels = []
        if not image_item.auto_levels:
            levels = image_item.levels.tolist()
        config = {"color_levels": levels}
        self.stateChanged.emit(config)

    # -----------------------------------------------------------------------
    # ROI methods

    @Slot(object)
    def _activate_roi_tool(self, roi_tool):
        if self._canvas is not None:
            self._deactivate_canvas()

        if roi_tool in [ROITool.NoROI, ROITool.Rect, ROITool.Crosshair]:
            self.roi.show(roi_tool)
        elif roi_tool in [ROITool.DrawRect, ROITool.DrawCrosshair]:
            # Hide shown ROIs to give way to the canvas
            self.roi.show(ROITool.NoROI)
            self._activate_canvas(roi_tool)

    def _activate_canvas(self, draw_tool):
        # Deactivate the canvas if a previous is applied
        if self._canvas is not None:
            self._deactivate_canvas()

        # Check if draw tool has an canvas counterpart
        if draw_tool not in ROI_CANVAS_MAP:
            return

        canvas_class, roi_tool = ROI_CANVAS_MAP[draw_tool]
        rect = self.plotItem.vb.viewRect()

        # Create and add canvas
        self._canvas = canvas_class(rect)
        self.plotItem.vb.addItem(self._canvas)
        self._canvas.editingFinished.connect(partial(self._draw_roi, roi_tool))

    @Slot(object)
    def _draw_roi(self, roi_tool, rect):
        self._deactivate_canvas()
        if rect.isValid():
            # Draw the ROI and show corresponding ROI tool
            new_rect = self.plotItem.mapRectFromTransform(rect)
            self.roi.add(roi_tool, new_rect.topLeft(), size=new_rect.size())
            self.roi.show(roi_tool)
            self.roi.selected.emit(roi_tool)
        else:
            # Do nothing and unselect the button
            self.roi.selected.emit(ROITool.NoROI)

    def _deactivate_canvas(self):
        self.plotItem.vb.removeItem(self._canvas)
        self._canvas.destroy()
        self._canvas = None

    # -----------------------------------------------------------------------
    # Scale legend

    def _show_scale_legend(self, show=True):
        if show:
            # Create the legend object
            if self._scale_legend is None:
                self._scale_legend = ScaleLegend()
                self._scale_legend.setParentItem(self.plotItem.vb)
                self._scale_legend.anchor(itemPos=(0, 1),
                                          parentPos=(0, 1),
                                          offset=(5, -5))
        else:
            # Destroy object
            if self._scale_legend is not None:
                self._scale_legend.setParentItem(None)
                self._scale_legend.deleteLater()
                self._scale_legend = None

    def _update_scale_legend(self):
        if self._scale_legend is not None:
            x_scale, y_scale = self.plotItem.axes_transform[TF_SCALING]
            x_units, y_units = (labels["units"] for labels in
                                self.plotItem.axes_labels)
            self._scale_legend.set_value(x_scale, y_scale, x_units, y_units)

    # -----------------------------------------------------------------------

    def event(self, event):
        if event.type() == QEvent.ToolTip:
            self.toolTipChanged.emit()
        return super().event(event)

    def sizeHint(self):
        """ The optimal size when all aux plots and labels are activated."""
        return QSize(610, 460)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import Constant

from karabogui.graph.common.formatting import table_body, table_row

from ..base.stats import BaseStats


class HistogramStats(BaseStats):

    # HTML table format for the stats
    html_table = Constant(
        "<b><u>Histogram Statistics</b></u><br/>"
        + table_row(table_body(header="Count", tabs=("{count}",)))
        + table_row(table_body(header="Mean", tabs=("{mean}",)))
        + table_row(table_body(header="Minimum", tabs=("{min}",)))
        + table_row(table_body(header="Maximum", tabs=("{max}",)))
        + table_row(table_body(header="Std. Dev.", tabs=("{std}",)))
        + table_row(table_body(header="Mode", tabs=("{mode}",)))
        + table_row(table_body(header="Bins", tabs=("{bins}",)))
        + table_row(table_body(header="Bin Width", tabs=("{bin_width}",)))
    )

    # -----------------------------------------------------------------------
    # Trait properties

    def _get_html(self):
        html = ''
        if self.stats:
            formatted_stats = {key: self._to_string(value)
                               for key, value in self.stats.items()}
            table = self.html_table.format(**formatted_stats)
            html = self.html_format.format(contents=table)
        return html
